@@ -1,7 +1,7 @@
 /*
     Wn: A Server for the HTTP
     File: wn/chkcntrl.c
-    Version 1.17.0
+    Version 1.18.3
 
     Copyright (C) 1995, 1996  <by John Franks>
 
@@ -28,12 +28,8 @@
 #include <sys/stat.h>
 #include "wn.h"
 #include "chkcntrl.h"
-#include "content.h"
 #include "access.h"
 
-#ifndef S_ISREG
-#define	S_ISREG(m)	(((m)&S_IFMT) == S_IFREG)
-#endif
 
 extern int	chkauth();
 
@@ -71,6 +67,8 @@ Request	*ip;
 	Cache_entry	*cep,
 			entry;
 
+
+
 #ifdef ACCESS_DENIED_URL
 	mystrncpy( dir_p->noaccess_url, ACCESS_DENIED_URL, MIDLEN/2);
 #else
@@ -93,33 +91,37 @@ Request	*ip;
 
 
 	if ( lstat( ip->cachepath, &stat_buf) != 0 ) {
-		writelog( ip,  ERRMSG2, ip->cachepath);
+		writelog( ip,  err_m[2], ip->cachepath);
 		ip->type = RTYPE_DENIED;
 		ip->status |= WN_CANT_STAT;
 		return;
 	}
 
+#ifndef S_ISREG
+#define	S_ISREG(m)	(((m)&S_IFMT) == S_IFREG)
+#endif
+
 	if ( !(S_ISREG(stat_buf.st_mode))) {
-		senderr( SERV_ERR, ERRMSG13, ip->cachepath);
-		wn_exit( 2);
+		senderr( SERV_ERR, err_m[13], ip->cachepath);
+		return;
 	}
 	ip->cache_uid = (unsigned) stat_buf.st_uid;
 	ip->cache_gid = (unsigned) stat_buf.st_gid;
 
 	if ( (!WN_OPT_U) && (WN_OPT_T) && (!is_trusted( ip)) ) {
-		logerr( ERRMSG7, ip->cachepath);
+		logerr( err_m[7], ip->cachepath);
 		ip->type = RTYPE_DENIED;
 		return;
 	}
 
 	if ( stat_buf.st_uid == (uid_t) USERID) {
-		logerr( ERRMSG8, ip->cachepath);
+		logerr( err_m[8], ip->cachepath);
 		ip->type = RTYPE_DENIED;
 		return;
 	}
 
 	if ( (fp = fopen( ip->cachepath, "r")) == (FILE *) NULL ) {
-		logerr( ERRMSG9, ip->cachepath);
+		logerr( err_m[9], ip->cachepath);
 		ip->type = RTYPE_DENIED;
 		return;
 	}
@@ -127,7 +129,7 @@ Request	*ip;
 
 	read_cache_dirinfo( fp, dir_p);
 
-	access_status = chkaccess( ip,  dir_p->accessfile);
+	access_status = chkaccess( ip->cachepath,  dir_p->accessfile);
 
 	switch( access_status) {
 	case ACCESS_GRANTED:
@@ -152,7 +154,7 @@ Request	*ip;
 
 	if ( *(dir_p->authmod) && (access_status != ACCESS_PRIVILEGED) ) {
 		if ( !(is_trusted( ip) || is_atrusted( ip)) ) {
-			senderr( DENYSTATUS, ERRMSG90, dir_p->authmod);
+			senderr( DENYSTATUS, err_m[90], dir_p->authmod);
 			ip->type = RTYPE_FINISHED;
 			fclose( fp);
 			return;
@@ -161,7 +163,7 @@ Request	*ip;
 		exec_ok( ip);
 
 		if ( !getfpath( dir_p->authmodule, dir_p->authmod, ip)) {
-			senderr( SERV_ERR, ERRMSG45, dir_p->authmod);
+			senderr( SERV_ERR, err_m[45], dir_p->authmod);
 			ip->type = RTYPE_FINISHED;
 			fclose( fp);
 			return;
@@ -181,8 +183,8 @@ Request	*ip;
 	if ( *(dir_p->filemod)) {
 		exec_ok( ip);
 		if ( !getfpath( dir_p->filemodule, dir_p->filemod, ip)) {
-			senderr( SERV_ERR, ERRMSG45, dir_p->filemod);
-			wn_exit( 2);
+			senderr( SERV_ERR, err_m[45], dir_p->filemod);
+			return; /* to process_url */
 		}
 		dir_p->filemod = dir_p->filemodule;
 		if ( WN_OPT_U )
@@ -192,19 +194,22 @@ Request	*ip;
 	if ( *(dir_p->cachemod)) {
 		exec_ok( ip);
 		if ( !getfpath( dir_p->cachemodule, dir_p->cachemod, ip)) {
-			senderr( SERV_ERR, ERRMSG45, dir_p->cachemod);
-			wn_exit( 2);
+			senderr( SERV_ERR, err_m[45], dir_p->cachemod);
+			return; /* to process_url */
 		}
 		dir_p->cachemod = dir_p->cachemodule;
 		if ( WN_OPT_U )
 			check_perm( ip, dir_p->cachemodule);
+
+		if ( (dir_p->defattributes) & WN_CGI)
+			cgi_env( ip, FALSE);
 	}
 
 	if ( *(dir_p->indexmod)) {
 		exec_ok( ip);
 		if ( !getfpath( dir_p->indexmodule, dir_p->indexmod, ip)) {
-			senderr( SERV_ERR, ERRMSG45, dir_p->indexmod);
-			wn_exit( 2);
+			senderr( SERV_ERR, err_m[45], dir_p->indexmod);
+			return;
 		}
 		dir_p->indexmod = dir_p->indexmodule;
 		if ( WN_OPT_U )
@@ -214,13 +219,15 @@ Request	*ip;
 	if ( iswndir( ip)) {
 		/* It's a title, keyword, grep, or index search of this dir */
 		fclose( fp);
-		ip->title = ip->encoding = ip->keywords = ip->filter = NULL;
+		ip->title = ip->encoding = ip->keywords 
+			= ip->filter = ip->handler = NULL;
 		ip->content_type = "text/html";
 		return;  
 	}
 
 	cep = &entry;
-	cep->line = ip->cacheline;
+	cep->line = ip->cacheline;  
+	/* actual cache line goes in ip struct not cep struct */
 
 	while ( read_cache_file( cep, fp, ip->basename)) {
 		if ( !*cep->basename)
@@ -228,48 +235,97 @@ Request	*ip;
 		if ( streq( ip->basename, cep->basename)) {
 			if ( ip->type == RTYPE_UNCHECKED)
 				ip->type = RTYPE_FILE;
+
+			ip->attributes |= 
+				(entry.attributes ? entry.attributes :
+						dir_p->defattributes);
+
 			ip->title = entry.title;
 			ip->content_type = entry.content; 
 			ip->encoding = entry.encoding; 
 			ip->keywords = entry.keywords; 
 
 			ip->includes = dir_p->defincludes;
+			ip->wrappers = dir_p->defwrapper;
+			ip->list_incl = dir_p->deflist;
+
 			if ( *entry.includes)
 				ip->includes = entry.includes;
 			if ( (*ip->includes) &&
-					!streq( ip->includes, "<none>"))
-				ip->attributes |= WN_INCLUDE;
+					!streq( ip->includes, "<none>")) {
+				ip->attrib2 |= WN_INCLUDE;
+				ip->attrib2 &= ~(WN_LIST_INCL);
+			}
 			else
-				*ip->includes = '\0';
+				ip->includes = empty;
 
-			ip->wrappers = dir_p->defwrapper;
 			if ( *entry.wrappers)
 				ip->wrappers = entry.wrappers; 
 			if ( (*ip->wrappers) &&
-					!streq( ip->wrappers, "<none>"))
-				ip->attributes |= WN_WRAPPED;
+					!streq( ip->wrappers, "<none>")) {
+				ip->attrib2 |= WN_WRAPPED;
+				ip->attrib2 &= ~(WN_LIST_INCL);
+			}
 			else
-				*ip->wrappers = '\0';
+				ip->wrappers = empty;
 
 			ip->inclptr = ip->wrappers; 
 
+			if ( *entry.list_incl)
+				ip->list_incl = entry.list_incl;
+
+			if ( (*ip->list_incl) &&
+					!streq( ip->list_incl, "<none>")) {
+				ip->attrib2 |= WN_LIST_INCL;
+				/* shut off any includes */
+				ip->attrib2 &= ~(WN_INCLUDE + WN_WRAPPED);
+				ip->includes = ip->wrappers = empty;
+			}
+			else
+				ip->list_incl = empty;
+
 			ip->swrapper = entry.swrapper; 
 			ip->nomatchsub = entry.nomatchsub; 
 			ip->filter = entry.filter;
+			ip->handler = entry.handler;
+
 			ip->expires = entry.expires;
 			ip->maxage = ((*entry.maxage) 
 				? entry.maxage : dir_p->default_maxage);
 
+			ip->handler = ((*entry.handler) 
+				? entry.handler : dir_p->def_handler);
+
+			if ( *ip->handler) 
+				if ( streq( ip->handler, "<none>")) {
+					ip->handler = empty;
+					ip->attributes &= ~WN_CGI;
+				}
+				else {
+					ip->attributes |= WN_CGI;
+					ip->type = RTYPE_CGI_HANDLER;
+				}
+
+			ip->filter = ((*entry.filter) 
+				? entry.filter : dir_p->def_filter);
+
+			if ( *ip->filter) 
+				if ( streq( ip->filter, "<none>")) {
+					ip->filter = empty;
+					ip->attributes &= ~WN_FILTERED;
+				}
+				else {
+					ip->attributes |= WN_FILTERED;
+				}
+
 			for ( i = 0; i < NUMFIELDS; i++) {
 				ip->field[i] = entry.field[i];
 			}
 
-			ip->attributes |= entry.attributes;
-			ip->attributes |= dir_p->defattributes;
-			/* One of these two attributes is always zero */
-
 			ip->filetype |= entry.filetype;
-			if ( ip->attributes & (WN_INCLUDE + WN_WRAPPED)) {
+
+			if ( ip->attrib2 & 
+				(WN_INCLUDE + WN_WRAPPED + WN_LIST_INCL)) {
 				update_mod_time( ip);
 				ip->attributes |= WN_PARSE;
 			}
@@ -279,26 +335,21 @@ Request	*ip;
 				ip->type = RTYPE_IMAGEMAP;
 			}
 
-			if ( *(entry.parse)){
-				if ( streq(entry.parse, "true"))
-					ip->attributes |= WN_PARSE;
-				else	/* it's false */
-					ip->attributes &= ~(WN_PARSE);
-			}
 			if ( ip->attributes & WN_NOPARSE)
 				ip->attributes &= ~(WN_PARSE);
 
-			/* By default CGI non-POST with no query is dynamic */
-			if ( ((ip->type == RTYPE_CGI ) ||
+			/* By default CGI is dynamic non-cachable */
+			if ( (ip->type == RTYPE_CGI ) ||
 						(ip->type == RTYPE_NPH_CGI ))
-					&& ( inheadp->method != POST)
-					&& ( !*(ip->query)))
-
-				ip->attributes |= WN_DYNAMIC;
+				ip->attributes |= (WN_DYNAMIC + WN_NOCACHE);
 
 			/* But it can be non-dynamic if dynamic = false */
 			if ( ip->attributes & WN_NONDYNAMIC)
 				ip->attributes &= ~(WN_DYNAMIC);
+
+			if ( ip->attributes & WN_CACHEABLE)
+				ip->attributes &= ~(WN_NOCACHE);
+
 			if ( *(entry.dynamic)){
 				if ( streq(entry.dynamic, "true"))
 					ip->attributes |= WN_DYNAMIC;
@@ -373,34 +424,35 @@ Request	*ip;
 	register char	*cp,
 			*cp2;
 
-	int		i = 0;
-
 	char		suffix[10];
 
 	ip->title = ip->basename;
 
+	if ( !*(dir_p->default_content))
+		dir_p->default_content = DEFAULT_CONTENT_TYPE;
+	ip->content_type = dir_p->default_content;
+	ip->maxage = dir_p->default_maxage;
+	ip->handler = dir_p->def_handler;
+	ip->filter = dir_p->def_filter;
+	ip->attributes |= dir_p->defattributes;
+
 	if ( *(dir_p->filemod))
 		/* skip check on file name */
 		;
 	else if ( (*(ip->title) == '.') || streq( ip->title, CACHEFNAME) ||
 				(ip->status & WN_CANT_STAT) ||
+				streq( ip->title, dir_p->accessfile) ||
 				streq( ip->title, CONTROLFILE_NAME)) {
 		ip->type = RTYPE_DENIED;
 		return;
 	}
 
-	if ( !*(dir_p->default_content))
-		dir_p->default_content = DEFAULT_CONTENT_TYPE;
-
 	if ( ip->type == RTYPE_UNCHECKED)
 		ip->type = RTYPE_FILE;
 
 	if ( (ip->type == RTYPE_CGI ) || (ip->type == RTYPE_NPH_CGI ))
 		ip->attributes |= WN_DYNAMIC;
 
-	ip->content_type = dir_p->default_content;
-	ip->maxage = dir_p->default_maxage;
-
 	if ( (cp = strrchr( ip->basename, '.')) == NULL )
 		return;
 
@@ -418,15 +470,7 @@ Request	*ip;
 		strlower( suffix);
 		*cp = '.';
 	}
-
-	while ( list[i][0] != NULL) {
-		if ( streq( list[i][0], suffix)) {
-			ip->content_type = list[i][1];
-			break;
-		}
-		i++;
-	}
-
+	get_mtype( suffix);
 	if ( strncasecmp( ip->content_type, "text", 4) == 0 )
 		ip->filetype |= WN_TEXT;
 	if ( streq( ip->content_type, "text/html"))
@@ -437,17 +481,17 @@ Request	*ip;
 
 
 /*
- * chkaccess( ip, accessfile) checks whether the client's IP address
+ * chkaccess( cachepath, accessfile) checks whether the client's IP address
  * is in the allowed list in accessfile.  Returns ACCESS_PRIVILEGED if 
- * access is unconditionally allowed, ACESS_GRANTED if further
+ * access is unconditionally allowed, ACCESS_GRANTED if further
  * authentication (through an authentication module) may be required, 
  * ACCESS_DENIED if access is denied and ACCESS_ERR on error.
  */
 
 int
-chkaccess( ip, accessfile)
-Request	*ip;
-char	*accessfile;
+chkaccess( cachepath, accessfile)
+char	*cachepath,
+	*accessfile;
 {
 	FILE	*fp;
 	int	len,
@@ -463,22 +507,19 @@ char	*accessfile;
 	if ( ! *accessfile)
 		return ACCESS_GRANTED;	/* No access control */
 
-	if ( getfpath2( buf, accessfile, ip) == FALSE)
+	if ( getfpath2( buf, accessfile, cachepath) == FALSE)
 		return ACCESS_ERR;
 
 	get_remote_info( );
 
-	if ( WN_OPT_U )
-		check_perm( ip, buf);
-
 	if ((fp = fopen( buf, "r")) == (FILE *)NULL ) {
-		logerr( ERRMSG87, buf);
+		logerr( err_m[87], buf);
 		return ACCESS_ERR;
 	}
 
 	while ( fgets( linebuf, SMALLLEN, fp)) {
 		if ( !chop( linebuf)) {
-			logerr( ERRMSG62, buf);
+			logerr( err_m[62], buf);
 			return ACCESS_ERR;
 		}
 
@@ -489,10 +530,10 @@ char	*accessfile;
 			continue;
 		}
 
-		if ( notflag = ( *cp == '!'))
+		if ( (notflag = ( *cp == '!')))
 			cp++;
 
-		if ( priv_flag = ( *cp == '+'))
+		if ( (priv_flag = ( *cp == '+')))
 			cp++;
 
 		len = strlen( cp);
@@ -563,9 +604,12 @@ char	*net,
 
 
 /*
- *  wild_match
+ *  wild_match( char *l, char *p)
  *
- *  String equality routine, including matching the '*' character.
+ *  String equality routine, including matching the '*' and '?' characters.
+ *  The string p contains the wildcards.  '?' matches any single character
+ *  while '*' matches any string.
+ *  
  */
 
 /*
@@ -574,27 +618,35 @@ char	*net,
  *
  */
 
-int wild_match(l,p)
-    char *l,
-         *p;
+int
+wild_match( l, p)
+char	*l,
+	*p;
 {
-    if (!*l) {
-        if (!*p) return(1);
-        else if (*p == '*') return(wild_match(l,p+1));
-        else return(0);
-        }
-    if (*p == '*') {
-        while (!wild_match(l,p+1)) {
-            l++;
-            if (!*l) {
-                if (!*(p+1)) return(1);
-                else return(0);
-                }
-            }
-        return(1);
-        }
-    if (*p == '?') return(wild_match(l+1,p+1));
-    return((*l == *p) && wild_match(l+1,p+1));
+	if (!*l) {
+		if (!*p)
+			return TRUE;
+		else if (*p == '*') 
+			return (wild_match( l, p+1));
+		else
+			return FALSE;
+	}
+	if (*p == '*') {
+		while ( !wild_match( l, p+1)) {
+			l++;
+			if (!*l) {
+				if (!*(p+1))
+					return TRUE;
+				else
+					return FALSE;
+				}
+			}
+		return TRUE;
+	}
+	if (*p == '?')
+		return(wild_match( l+1, p+1));
+
+	return ((*l == *p) && wild_match( l+1, p+1));
 }
 
 
@@ -613,7 +665,7 @@ Dir_info	*dep;
 {
 	register char	*cp;
 	char		*field,
-			*value;
+			*value = NULL;
 
 	cp = dep->dirline;
 	if ( fgets( cp, BIGLEN, fp) == NULL) {
@@ -622,7 +674,7 @@ Dir_info	*dep;
 	}
 
 	if ( strrchr( dep->dirline, '\n') == NULL) {
-		senderr( SERV_ERR, ERRMSG63, "");
+		senderr( SERV_ERR, err_m[63], "");
 		wn_exit( 2);
 	}
 
@@ -632,10 +684,12 @@ Dir_info	*dep;
 		cp++;
 
 	dep->accessfile = dep->swrapper = dep->defwrapper = dep->defincludes
-	= dep->subdirs = dep->nomatchsub = dep->dir_owner 
+	= dep->deflist = dep->subdirs = dep->nomatchsub = dep->dir_owner 
 	= dep->cachemod = dep->filemod = dep->indexmod
-	= dep->authmod = dep->authtype = dep->authrealm 
-	= dep->default_content = dep->defdoc = dep->default_maxage = cp;
+	= dep->authmod = dep->authtype = dep->authrealm = cp;
+
+	dep->default_content = dep->defdoc = dep->default_maxage
+	= dep->def_filter = dep->def_handler = cp;
 
 	*dep->authmodule = *dep->cachemodule
 	 = *dep->filemodule = *dep->indexmodule = '\0';
@@ -656,12 +710,12 @@ Dir_info	*dep;
 			if ( *(cp-1) == '\\')	/* ignore escaped & */
 				break;
 			*cp = '\0';
-			setdirvalue( field, value, dep);
+			setdirvalue( field, value, cp, dep);
 			field = ++cp;
 			break;
 		case '\n':
 			*cp = '\0';
-			setdirvalue( field, value, dep);
+			setdirvalue( field, value, cp, dep);
 			return;
 		default:
 			cp++;
@@ -671,15 +725,25 @@ Dir_info	*dep;
 
 
 static void
-setdirvalue( field, value, dep)
+setdirvalue( field, value, end, dep)
 char		*field,
-		*value;
+		*value,
+		*end;
 Dir_info	*dep;
 {
-	if ( !*value) {
-		logerr( ERRMSG10, field);
+	char	buf[TINYLEN];
+
+	if ( !value || !*value) {
+		mystrncpy( buf, field, TINYLEN);
+		logerr( err_m[10], buf);
 		return;
 	}
+	if ( (end - value) > MIDLEN/2) {
+		mystrncpy( buf, field, TINYLEN);
+		logerr( err_m[107], buf);
+		value[MIDLEN/2] = '\0';
+	}
+
 	switch (*field) {
 	case 'a':
 		if ( strncmp( field, "auth", 4) == 0) {
@@ -722,6 +786,19 @@ Dir_info	*dep;
 			dep->defincludes = value;
 		}
 
+		else if ( streq( field, "default_filter")) {
+			dep->def_filter = value;
+		}
+
+		else if ( streq( field, "default_handler")) {
+			dep->def_handler = value;
+			dep->attributes |= WN_CGI;
+		}
+
+		else if ( streq( field, "deflist")) {
+			dep->deflist = value;
+		}
+
 		else if ( streq( field, "defattributes"))
 			dep->defattributes = (unsigned) atol( value);
 
@@ -735,7 +812,7 @@ Dir_info	*dep;
 				dep->default_maxage = value;
 
 		else
-			logerr( ERRMSG11, field);
+			logerr( err_m[11], field);
 		break;
 
 	case 'f':	/* file module */
@@ -767,7 +844,7 @@ Dir_info	*dep;
 			break;
 
 		default:
-			logerr( ERRMSG11, field);
+			logerr( err_m[11], field);
 			break;
 		}
 		break;
@@ -786,7 +863,7 @@ Dir_info	*dep;
 		break;
 		
 	default:
-		logerr( ERRMSG11, field);
+		logerr( err_m[11], field);
 	}
 }
 
@@ -825,12 +902,12 @@ char		*key;
 		}
 
 		if ((lfp = popen( dir_p->cachemod, "r")) == (FILE *)NULL ) {
-			senderr( SERV_ERR, ERRMSG40, dir_p->cachemod);
+			senderr( SERV_ERR, err_m[40], dir_p->cachemod);
 			wn_exit( 2);
 		}
 
 		if ( (c = getc( lfp)) == EOF ) {
-			senderr( SERV_ERR, ERRMSG43, dir_p->cachemod);
+			senderr( SERV_ERR, err_m[43], dir_p->cachemod);
 			pclose( lfp);
 			wn_exit( 2);
 		}
@@ -853,17 +930,18 @@ char		*key;
 		pclose( lfp);
 
 	if ( strrchr( cp, '\n') == NULL) {
-		senderr( SERV_ERR, ERRMSG63, "");
+		senderr( SERV_ERR, err_m[63], "");
 		wn_exit( 2);
 	}
 
 	while ( *cp)
 		cp++;
 	cep->end = cep->basename = cep->title = cep->keywords = cep->dynamic
-	= cep->content = cep->encoding = cep->includes = cep->cgi
-	= cep->wrappers = cep->swrapper = cep->nomatchsub = cep->parse
-	= cep->filter = cep->expires = cep->maxage = cep->url 
-	= cep->redirect = cp;
+	= cep->content = cep->encoding = cep->includes = cep->list_incl 
+	= cep->cgi = cep->wrappers = cep->swrapper = cep->nomatchsub
+	= cep->filter = cep->handler = cep->expires = cep->maxage
+	= cep->url = cep->redirect = cep->handler = cp;
+
 	for ( i = 0; i < NUMFIELDS; i++) {
 		cep->field[i] = cp;
 	}
@@ -886,7 +964,7 @@ char		*key;
 				break;
 			}
 			*cp = '\0';
-			setvalue( field, value, cep);
+			setvalue( field, value, cp, cep);
 			field = ++cp;
 			if ( (value = strchr( cp, '=' )) == NULL )
 				cp = value = cep->end;
@@ -897,7 +975,7 @@ char		*key;
 			break;
 		case '\n':
 			*cp = '\0';
-			setvalue( field, value, cep);
+			setvalue( field, value, cp, cep);
 			return TRUE;
 		default:
 			cp++;
@@ -907,26 +985,31 @@ char		*key;
 }
 
 static void
-setvalue( field, value, cep)
+setvalue( field, value, end, cep)
 char	*field,
-	*value;
+	*value,
+	*end;
 Cache_entry	*cep;
 {
 	char	buf[SMALLLEN];
 	int	i;
 
 	if ( !*value) {
-		strcpy( buf, field);
+		mystrncpy( buf, field, 28);
 		strcat( buf, ":");
 		mystrncat( buf, cep->basename, SMALLLEN - 30);
-		logerr( ERRMSG5, buf);
+		logerr( err_m[5], buf);
 		return;
 	}
+	if ( (end - value) > MIDLEN/2) {
+		mystrncpy( buf, field,  30);
+		logerr( err_m[107], buf);
+		value[MIDLEN/2] = '\0';
+	}
+
 	switch (*field) {
 	case 'a':		/* attributes */
 		cep->attributes |= (unsigned) atol( value);
-		dir_p->defattributes = 0;
-		/* Explicit attrib given, don't use default */
 		break;
 	case 'c':
 		if ( field[1] == 'g' )		/* cgi */
@@ -935,8 +1018,6 @@ Cache_entry	*cep;
 			cep->content = value;
 		break;
 
-
-		break;
 	case 'd':
 		cep->dynamic = value;
 		break;
@@ -958,27 +1039,33 @@ Cache_entry	*cep;
 			break;
 		case 'l':		/* field */
 			i = atoi( field + 5);
-			cep->field[i] = value;
+			if ( (i >= 0 ) && ( i < NUMFIELDS))
+				cep->field[i] = value;
 			break;
 		case 't':		/* filter */
 			cep->filter = value;
-			cep->attributes |= WN_FILTERED;
 			break;
 		}
 		break;
 	case 'h':
-		if ( strlen( cep->headerlines) + strlen( value) > BIGLEN ) {
-			senderr( SERV_ERR, ERRMSG70, value);
-			wn_exit( 2);
+		if ( streq( field, "handler"))
+			cep->handler = value;
+
+		else if ( streq( field, "header")) {
+			if ( strlen( cep->headerlines) + 
+						strlen( value) > BIGLEN ) {
+				senderr( SERV_ERR, err_m[70], value);
+				wn_exit( 2);
+			}
+			strcat( cep->headerlines, value);
+			strcat( cep->headerlines, "\n");
 		}
-		strcat( cep->headerlines, value);
-		strcat( cep->headerlines, "\n");
 		break;
 	case 'i':
 		switch( field[1]) {
 		case 'n':
 			cep->includes = value;
-			cep->attributes |= WN_INCLUDE;
+			cep->attrib2 |= WN_INCLUDE;
 			break;
 		case 'm':
 			if ( streq( value, "true"))
@@ -989,6 +1076,10 @@ Cache_entry	*cep;
 	case 'k':
 		cep->keywords = value;
 		break;
+	case 'l':			/* list */
+		cep->list_incl = value;
+		cep->attrib2 |= WN_LIST_INCL;
+		break;
 	case 'm':
 		cep->maxage = value;
 		break;
@@ -998,9 +1089,6 @@ Cache_entry	*cep;
 		else if ( streq( value, "true")) /* nosearch=true */
 			cep->attributes |= WN_NOSEARCH;
 		break;
-	case 'p':
-		cep->parse = value;
-		break;
 	case 'r':
 		cep->redirect = value;
 		break;
@@ -1017,10 +1105,10 @@ Cache_entry	*cep;
 		break;
 	case 'w':
 		cep->wrappers = value;
-		cep->attributes |= WN_WRAPPED;
+		cep->attrib2 |= WN_WRAPPED;
 		break;
 	default:
-		logerr( ERRMSG6, field);
+		logerr( err_m[6], field);
 	}
 }
 
@@ -1043,8 +1131,16 @@ char	*buf;
 
 	struct stat stat_buf;
 
-	if ( (!WN_OPT_T) && (!WN_OPT_U) )  /* No restrictions */
-		return;
+	if ( (!WN_OPT_T) && (!WN_OPT_U) ) {
+		if ( (ip->filetype & WN_NOT_WORLD_READ)
+				&& (ip->filetype & WN_NOT_REG_FILE)) {
+			senderr( SERV_ERR, err_m[44], buf);
+			wn_exit( 2); /* senderr: SERV_ERR */
+		}
+		else
+			return;
+
+	}
 	else if ( is_trusted( ip) )
 		return;
 
@@ -1055,7 +1151,7 @@ char	*buf;
 			cp++;
 		*cp = '\0';
 		if ( stat( filebuf, &stat_buf) != 0 ) {
-			logerr( ERRMSG12, buf);
+			logerr( err_m[12], buf);
 			return;
 		}
 
@@ -1065,7 +1161,7 @@ char	*buf;
 			return;
 	}
 		
-	senderr( DENYSTATUS, ERRMSG3, ip->cachepath);
+	senderr( DENYSTATUS, err_m[3], filebuf);
 	wn_exit( 2);
 }
 
@@ -1076,12 +1172,12 @@ Request	*ip;
 {
 	
 #ifdef FORBID_CGI
-	senderr( DENYSTATUS, ERRMSG4, ip->cachepath);
+	senderr( DENYSTATUS, err_m[4], ip->cachepath);
 	wn_exit( 2);
 #else
 	if ((serv_perm & WN_FORBID_EXEC) ||
 			((serv_perm & WN_RESTRICT_EXEC) && !is_trusted( ip))) {
-		senderr( DENYSTATUS, ERRMSG4, ip->cachepath);
+		senderr( DENYSTATUS, err_m[4], ip->cachepath);
 		wn_exit( 2);
 	}
 #endif
