@@ -1,7 +1,7 @@
 /*
     Wn: A Server for the HTTP
     File: wn/common.h
-    Version 1.15.7
+    Version 1.18.7
     
     Copyright (C) 1996  <by John Franks>
 
@@ -31,20 +31,20 @@
 
 #define STANDARD_PORT	(80)
 
-#define	BIGLEN		(4096)
-#define	CACHELINE_LEN	(2*BIGLEN)
 #define	MIDLEN		(2048)
+#define	BIGLEN		(2*MIDLEN)
+#define	CACHELINE_LEN	(4*MIDLEN)
 #define	RANGELEN	(1024)
 #define	SMALLLEN	(256)
 #define	TINYLEN		(32)
 #define INBUFFSIZE	(2048)
-#define OUT_BUFFSIZE	(128)
+#define OUT_BUFFSIZE	(4096)
 
 /* Bits in the Request attributes */
 #define WN_DYNAMIC	(1)
 #define WN_NONDYNAMIC	(2)
-#define	WN_INCLUDE	(4)
-#define	WN_WRAPPED	(8)
+#define WN_POST_OK	(4)
+
 #define	WN_SWRAPPED	(16)
 #define	WN_FILTERED	(32)
 #define	WN_NOSEARCH	(64)
@@ -54,6 +54,14 @@
 #define WN_ISMAP	(1024)
 #define WN_NOCACHE	(2048)
 #define WN_UNBUFFERED	(4096)
+#define WN_CACHEABLE	(8192)
+
+
+/* Bits in the request attrib2 */
+#define	WN_INCLUDE		(1)
+#define	WN_WRAPPED		(2)
+#define	WN_LIST_INCL		(4)
+#define	WN_ISSEARCH		(8)
 
 
 
