@@ -1,9 +1,9 @@
 /*
     Wn: A Server for the HTTP
     File: wn/init.c
-    Version 1.17.0
+    Version 1.18.7
     
-    Copyright (C) 1995, 1996  <by John Franks>
+    Copyright (C) 1995-8  <by John Franks>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -25,6 +25,7 @@
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/file.h>
+#include <sys/stat.h>
 #include <sys/signal.h>
 #include <string.h>
 #include <errno.h>
@@ -41,6 +42,8 @@
 #include <fcntl.h>
 #include "wn.h"
 #include "version.h"
+#include "err.h"
+#include "content.h"
 
 #ifdef USE_VIRTUAL_HOSTS
 extern char	*vhostlist[][3];
@@ -64,6 +67,7 @@ extern long	timezone;
 #endif
 
 extern char	*optarg;
+
 extern int	optind;
 
 #define		WN_COMMON	(0)
@@ -76,16 +80,22 @@ static int	log_type = WN_VERBOSE;
 static int	log_type = WN_COMMON;
 #endif
 
+
 static void	dump_log();
 
 char	rootdir[SMALLLEN],
+	wn_tmpdir[sizeof(TEMPDIR) + 20] = {TEMPDIR},
 	wnlogfile[SMALLLEN],
 	errlogfile[SMALLLEN],
 	pid_file[SMALLLEN],
 	cfname[SMALLLEN],
+	empty[1],	/* safe place to park pointers */
 	hostname[MAXHOSTNAMELEN],
 	remotehost[MAXHOSTNAMELEN],
-	remaddr[20];
+	remaddr[20],
+	**mtypelist = NULL,
+	**suflist = NULL;
+
 
 int	admin_mode = FALSE,
 	debug_log = FALSE;
@@ -108,7 +118,7 @@ static void	locklog(),
 
 static void	log_logfile(),
 		start_log(),
-		reopen_wnlog();
+		restart();
 
 
 static FILE	*logfp = NULL,
@@ -147,6 +157,7 @@ char	*argv[];
 #endif
 #endif
 
+	empty[0] = '\0';
 	port = DEFAULT_PORT;
 	mystrncpy (wnlogfile, WN_LOGFILE, SMALLLEN);
 	mystrncpy (errlogfile, WN_ERRLOGFILE, SMALLLEN);
@@ -190,7 +201,8 @@ char	*argv[];
 				break;
 			case 'L':
 				mystrncpy( wnlogfile, optarg, SMALLLEN);
-				log_func = LFILE;
+				if ( *wnlogfile)
+					log_func = LFILE;
 				break;
 			case 'l':
 				mystrncpy( errlogfile, optarg, SMALLLEN);
@@ -243,10 +255,12 @@ char	*argv[];
 				fprintf( stderr, "Unknown log type %s\n", 
 					optarg);
 				break;
+#ifdef USE_VIRTUAL_HOSTS
 #ifdef VIRTUAL_HOSTS_FILE
 			case 'V':
 				mystrncpy( vhostfile, optarg, SMALLLEN);
 				break;
+#endif
 #endif
 			case 'd':
 				/*should only be used for logging to file*/
@@ -300,20 +314,58 @@ char	*argv[];
 	open_wnlog( wnlogfile, errlogfile);
 #endif
 
+
+
+
 #ifdef USE_VIRTUAL_HOSTS
 #ifdef VIRTUAL_HOSTS_FILE
 	load_virtual();
 #endif
 #endif
 
+#ifdef STANDALONE
+	init_mime();
+#endif
+
 	WN_EXTRA_INIT
 }
 
+/*
+ * void open_wnlog( logfile, errlog) is called once on startup to 
+ * open the logfile, the error log file and to create a sub temp directory.
+ * The point of the sub temp directory is to make temp files more 
+ * secure.  If the server can't create this temp directory with 
+ * mode 0700 or if there isn't already one their owned by the server
+ * uid, then the server won't run for security reasons.  Unfortunately,
+ * this tmp directory never gets removed when the server is shut down.
+ */
+
 void
 open_wnlog( logfile, errlog)
 char	*logfile,
 	*errlog;
 {
+#ifndef FORBID_CGI
+	char		buf[TINYLEN];
+	struct stat	stat_buf;
+	uid_t		my_id;
+
+	my_id = getuid();
+	sprintf( buf, "/wn_tmp%d", my_id);
+	strcat( wn_tmpdir, buf);
+	if ( lstat( wn_tmpdir, &stat_buf) == -1) {
+		if ( (errno != ENOENT) || (mkdir( wn_tmpdir, 0700) != 0) ) {
+			fprintf( stderr, err_m[0], wn_tmpdir);
+			exit( 2);
+		}
+	}
+	else if ( (!S_ISDIR(stat_buf.st_mode))
+				|| (stat_buf.st_uid != my_id)
+				|| (chmod( wn_tmpdir, 0700) != 0) ) {
+		fprintf( stderr, err_m[0], wn_tmpdir);
+		exit( 2);
+	}
+#endif
 	if ( *logfile != '\0')
 		if ( (logfp = fopen( logfile, "a")) == NULL ) {
 #ifdef STANDALONE
@@ -323,8 +375,11 @@ char	*logfile,
 #endif
 			exit( 2);
 		}
-	if ( *errlog != '\0')
-		if ( (errlogfp = fopen( errlog, "a")) == NULL ) {
+	if ( *errlog != '\0') {
+		if ( streq( errlog, logfile))
+			errlogfp = logfp;
+
+		else if ( (errlogfp = fopen( errlog, "a")) == NULL ) {
 #ifdef STANDALONE
 			fprintf( stderr, 
 				"Can't open error logfile: %s\n", errlog);
@@ -333,16 +388,17 @@ char	*logfile,
 #endif
 			exit( 2);
 		}
+	}
 #ifdef STANDALONE
 	start_log( FALSE);
-	signal( SIGHUP, reopen_wnlog);
+	signal( SIGHUP, restart);
 #endif
 }
 
 
 
 static void
-reopen_wnlog( )
+restart( )
 {
 
 #ifdef STANDALONE
@@ -362,7 +418,9 @@ reopen_wnlog( )
 	}
 
 	if ( *errlogfile) {
-		if ( (errlogfp = 
+		if ( streq( errlogfile, wnlogfile))
+			errlogfp = logfp;
+		else if ( (errlogfp = 
 				freopen( errlogfile, "a", errlogfp)) == NULL ){
 			fprintf( stderr, "Can't reopen error logfile: %s\n",
 					errlogfile);
@@ -371,9 +429,9 @@ reopen_wnlog( )
 	}
 
 	start_log( TRUE);
-
-	signal( SIGHUP, reopen_wnlog);
-#endif
+	init_mime();
+	signal( SIGHUP, restart);
+#endif /* STANDALONE */
 }
 
 
@@ -398,6 +456,9 @@ char	*msg,
 	if ((log_func == LNOLOG) && !*errlogfile)
 		return;
 
+	if ( msg2 == NULL)
+		msg2 = "";
+
 	get_remote_info( );
 
 #ifndef NO_SYSLOG
@@ -444,8 +505,9 @@ char	*msg,
 }
 
 void
-daemon_logerr(  msg, error)
-char	*msg;
+daemon_logerr(  msg, msg2, error)
+char	*msg,
+	*msg2;
 int	error;
 {
 
@@ -466,8 +528,9 @@ int	error;
 
 #ifndef NO_SYSLOG
 	if ( (log_func == LSYSLOG) && !*errlogfile) {
-		sprintf( buf, "none - - [] \"none\" 500 0  <%s: errno=%d> ",
-				msg, error);
+		sprintf( buf, 
+			"none - - [] \"none\" 500 0  <%s: %.100s errno=%d> ",
+				msg, msg2, error);
 
 		if ( log_type == WN_VERBOSE ) {
 #ifdef USE_VIRTUAL_HOSTS
@@ -502,7 +565,8 @@ int	error;
 	fseek( errfp, 0L, 2);   
 	fprintf(errfp, "none - - [%s] \"none\" 500 0",  date);
 	if ( log_type == WN_VERBOSE ) {
-		fprintf( errfp, " <%s: errno=%d> <> <> <>", msg, error);
+		fprintf( errfp, " <%s %.100s: errno=%d> <> <> <>",
+					msg, msg2, error);
 #ifdef USE_VIRTUAL_HOSTS
 		fprintf( errfp, " <0>");
 #endif
@@ -557,10 +621,11 @@ char	*msg,
 	register char	*cp,
 			*cp2;
 
-	char	bytes[TINYLEN],
+	char	*authname,
+		bytes[TINYLEN],
 		status[TINYLEN],
 		date[TINYLEN],
-		lbuf[BIGLEN];
+		lbuf[2*SMALLLEN];
 
 	if ( outheadp && (*outheadp->status)) {
 		mystrncpy( status, outheadp->status, 4);
@@ -568,10 +633,7 @@ char	*msg,
 	else
 		strcpy( status, "200");
 
-	if ( *(ip->length))
-		strcpy( bytes, ip->length);
-	else
-		strcpy( bytes, "-");
+	sprintf( bytes,"%ld", ip->logcount);
 
 	time(&clock);
 	ltm = (struct tm *) localtime(&clock);
@@ -580,7 +642,8 @@ char	*msg,
 
 	cp = cp2 = this_conp->log_ptr;
 
-	sprintf( lbuf, "[%s] \"%s\" %s %s",
+	authname = ( *this_rp->authuser ? this_rp->authuser : "-");
+	sprintf( lbuf, "%s [%s] \"%.200s\" %s %s", authname,
 				date, ip->request, status, bytes);
 	cp2 += strlen( lbuf);
 	if ( cp2 >= this_conp->logbuf + LOGBUFLEN) {
@@ -599,10 +662,11 @@ char	*msg,
 			cp = cp2 = this_conp->logbuf;
 		}
 
-		strcpy( cp , lbuf);
+ 		
+		strcpy( cp, lbuf);
 		cp = cp2;
 
-		sprintf( lbuf, " <%s>", inheadp->ua);
+		sprintf( lbuf, " <%.500s>", inheadp->ua);
 		cp2 += strlen( lbuf);
 		if ( cp2 >= this_conp->logbuf + LOGBUFLEN) {
 			dump_log();
@@ -612,7 +676,7 @@ char	*msg,
 		strcpy( cp , lbuf);
 		cp = cp2;
 
-		sprintf( lbuf, " <%s>", inheadp->referrer);
+		sprintf( lbuf, " <%.500s>", inheadp->referrer);
 		cp2 += strlen( lbuf);
 		if ( cp2 >= this_conp->logbuf + LOGBUFLEN) {
 			dump_log();
@@ -621,7 +685,7 @@ char	*msg,
 		strcpy( cp , lbuf);
 		cp = cp2;
 
-		sprintf( lbuf, " <%s>", inheadp->cookie);
+		sprintf( lbuf, " <%.500s>", inheadp->cookie);
 		cp2 += strlen( lbuf);
 		if ( cp2 >= this_conp->logbuf + (LOGBUFLEN - 32)) {
 			dump_log();
@@ -706,17 +770,14 @@ char	*msg,
 	else
 		strcpy( status, "200");
 
-	if ( *(ip->length))
-		strcpy( bytes, ip->length);
-	else
-		strcpy( bytes, "-");
-
+	sprintf( bytes,"%ld", ip->logcount);
 
 	rfc931p = ( *this_conp->rfc931name ? this_conp->rfc931name : "-");
-	authname = ( *this_conp->authuser ? this_conp->authuser : "-");
+	authname = ( *ip->authuser ? ip->authuser : "-");
 
-	rhost = ( *remotehost ? remotehost : remaddr);
-	rhost = ( *rhost ? rhost : "unknown");
+	get_remote_info();
+
+	rhost = remotehost;
 
 	syslog( priority,"%s %s %s [] \"%s\" %s %s  <%s: %s>\n", 
 		rhost, rfc931p, authname, ip->request, status, 
@@ -784,6 +845,15 @@ int	restart;
 }
 
 
+void
+wn_abort( )
+{
+
+	this_rp->type = RTYPE_FINISHED;
+	this_rp->status |= (WN_ABORTED + WN_ERROR);
+	
+}
+
 void
 wn_exit( status)
 int	status;
@@ -806,7 +876,6 @@ dump_log()
 			*cp2;
 
 	char		*rfc931p,
-			*authname,
 			*rhost,
 			logline[MIDLEN];
 
@@ -830,8 +899,6 @@ dump_log()
 #endif
 
 	rfc931p = ( *this_conp->rfc931name ? this_conp->rfc931name : "-");
-	authname = ( *this_conp->authuser ? this_conp->authuser : "-");
-
 	rhost = ( *remotehost ? remotehost : remaddr);
 	rhost = ( *rhost ? rhost : "unknown");
 
@@ -844,8 +911,7 @@ dump_log()
 		*cp2 = *cp;
 		if ( *cp2 == '\n') {
 			*++cp2 = '\0';
-			fprintf( logfp, "%s %s %s %s", rhost, rfc931p,
-					authname, logline);
+			fprintf( logfp, "%s %s %s", rhost, rfc931p, logline);
 			cp2 = logline;
 			cp++;
 			continue;
@@ -861,5 +927,48 @@ dump_log()
 #endif
 
 }
+void
+get_mtype( suffix )
+char	*suffix;
+{
+	int	i,
+		unknown_mime = TRUE;
 
+	for ( i = 0; mimelist[i][0] != NULL; i++) {
+		if ( streq( mimelist[i][0], suffix)) {
+			mystrncpy( this_rp->contype, mimelist[i][1], SMALLLEN);
+			this_rp->content_type = this_rp->contype;
+			unknown_mime = FALSE;
+			break;
+		}
+	}
 
+#ifdef STANDALONE
+	if ( unknown_mime && (suflist != NULL)) {
+		for ( i = 0; suflist[i] != NULL; i++) {
+			if ( streq( suflist[i], suffix)) {
+				mystrncpy( this_rp->contype, 
+						mtypelist[i], SMALLLEN);
+				this_rp->content_type = this_rp->contype;
+				break;
+			}
+		}
+	}
+#endif  /* STANDALONE */
+}
+
+#ifdef STANDALONE
+int
+addsuffix( suf)
+char	*suf;
+
+{
+	int	i;
+
+	for ( i = 0; mimelist[i][0] != NULL; i++) {
+		if ( streq( suf, mimelist[i][0]))
+			return FALSE;
+	}
+	return TRUE;
+}
+#endif  /* STANDALONE */
