@@ -1,9 +1,9 @@
 /*
     Wn: A Server for the HTTP
     File: wn/parse.c
-    Version 1.15.3
+    Version 1.18.7
     
-    Copyright (C) 1996  <by John Franks>
+    Copyright (C) 1996, 1997  <by John Franks>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -47,8 +47,8 @@ static char	*in_anchor();
 static FILE	*openfp();
 
 
-static int	ifcnt = 0,
-		hidden_ifcnt = 0;
+static int	isinlist(),
+		scnt = 0;
 
 static Parsedata	swpdata;
 
@@ -61,6 +61,23 @@ typedef struct File_or_Pipe {
 static File_or_Pipe	swrapfp_st = { NULL, FRP_FILE};
 
 
+WN_CONST 
+char * WN_CONST parserr_m[] = {
+	/* 0 */ "",
+	/* 1 */ "Parse error: 'else' or 'elif' with no 'if'",
+	/* 2 */ "Parse error: 'endif' with no 'if'",
+	/* 3 */ "Parse error: 'start' inside 'if' construct or multiple starts with no 'end'",
+	/* 4 */ "Parse error: 'end' inside 'if' construct or multiple ends with no 'start'",
+	/* 5 */ "Parse line not understood",
+	/* 6 */ "Parse error: Bad if accessfile format",
+	/* 7 */ "Parse error: Bad if - regexp format",
+	/* 8 */ "Conditional text file line overflow",
+	/* 9 */ "Can't open Conditional text file",
+	/* 10 */ "Redirect attempted after text sent",
+	/* 11 */ "Missing quote mark in redirect URL",
+	/* 12 */ "If-Elif-Else nested too deeply"
+};
+
 /*
  * do_wrap( ip, show)
  * Handle inserted files or files with wrappers.  
@@ -96,23 +113,28 @@ unsigned	show;
 		wrap_place = before_file;
 		toplevel = TRUE;
 		ip->do_wrap_1st_time = FALSE;
-		ifcnt = 0;
-		hidden_ifcnt = 0;
+		scnt = 0;
 	}
 
+
 	cp = ip->inclptr;
-	if ( !*cp ) {		/* end of wrappers send the file */
+
+	if ( !*cp ) {
 		pdata.currfile = ip->basename;
 		switch ( wrap_place) {
 		case before_file:
-			wrap_place = in_file;
-			ip->inclptr = ip->includes;
-
+			/* end of wrappers send the file */
 			if ( !(ip->filetype & WN_TEXT)) {
-				senderr( DENYSTATUS, ERRMSG34,
+				senderr( DENYSTATUS, err_m[34],
 					ip->relpath);
 				wn_exit( 0);
 			}
+
+			wrap_place = in_file;
+
+			ip->inclptr = ip->includes;
+			/* this will be overridden for list include */
+
 			send_fp( ip, &pdata);
 			wrap_place = after_file;
 			/* If some includes are left, tack them on the end */
@@ -123,7 +145,7 @@ unsigned	show;
 		case in_file:
 		case after_file:
 			if ( !*cp ) {
-				logerr( ERRMSG35, "");
+				logerr( err_m[35], "");
 				return;
 			}
 		}
@@ -137,11 +159,19 @@ unsigned	show;
 	pdata.currfile = ip->inclptr;
 	if ( (openfp( ip->inclptr, ip, &frp_st)) == (FILE *) NULL ) {
 		/* Error logged in openfp */
-		ip->inclptr = cp;
+
+		if ( ip->attrib2 & WN_LIST_INCL)
+			ip->inclptr = ip->includes; /* safe empty spot */
+		else
+			ip->inclptr = cp;
 		return;
 	}
 	else {
-		ip->inclptr = cp;
+		if ( ip->attrib2 & WN_LIST_INCL)
+			ip->inclptr = ip->includes; /* safe empty spot */
+		else
+			ip->inclptr = cp;
+
 		send_include( ip, frp_st, &pdata);
 
 		switch ( wrap_place) {
@@ -180,6 +210,12 @@ Parsedata	*pdp;
 	while ( fgets( linebuf, BIGLEN, frp_st.fp)) {
 		pdp->currline++;
 		token = get_parse_token( linebuf, insert_param, pdp);
+
+		/* ignore if using section but not between start and end */
+		if ( (pdp->show & (SECT_MSK)) && !(pdp->show & (IN_SECT_MSK))
+						&& ( token != START))
+			continue;
+
 		if ( ( token == NULLTOKEN) || (token == COMMENT)) {
 			if ( (pdp->show & SHOW_IT) == SHOW_IT)
 				send_text_line( linebuf);
@@ -222,6 +258,12 @@ Parsedata	*pdp;
 	while ( fgets( linebuf, BIGLEN, ip->fp)) {
 		pdp->currline++;
 		token = get_parse_token( linebuf, insert_param, pdp);
+
+		/* ignore if using section but not between start and end */
+		if ( (pdp->show & (SECT_MSK)) && !(pdp->show & (IN_SECT_MSK))
+						&& ( token != START))
+			continue;
+
 		if ( ( token == NULLTOKEN) || (token == COMMENT)) {
 			if ( (pdp->show & SHOW_IT) == SHOW_IT)
 				send_text_line( linebuf);
@@ -276,10 +318,21 @@ char		*insert_ptr;
 	case INCLUDE:				/* include */
 		do_wrap( ip, pdp->show);
 		break;
+
+	case LIST_INCLUDE:			/* list include */
+		if ( isinlist( insert_ptr, ip->list_incl)) {
+			ip->inclptr = insert_ptr;
+			do_wrap( ip, pdp->show);
+		}
+		else
+			logerr( err_m[111], insert_ptr);
+
+		break;
+
 	case REDIRECT:
 		if ( (pdp->show & SHOW_IT) == SHOW_IT ) {
 			if ( ip->status & WN_PROLOGSENT ) {
-				logerr( PARSE_ERRMSG10, pdp->currfile);
+				logerr( parserr_m[10], pdp->currfile);
 				break;
 			}
 			dolocation( outheadp->redirect, ip);
@@ -293,6 +346,20 @@ char		*insert_ptr;
 		}
 		do_wrap( ip, show2);
 		break;
+	case LIST_SECTION:
+		if ( !isinlist( insert_ptr, ip->list_incl)) {
+			logerr( err_m[111], insert_ptr);
+			break;
+		}
+		show2 = pdp->show;
+		if ( !(show2 & SECT_MSK)) {
+			show2 |= SECT_MSK;
+			show2 &= ~IN_SECT_MSK;
+		}
+		ip->inclptr = insert_ptr;
+		do_wrap( ip, show2);
+		break;
+
 	case QUERY:
 		if ( (pdp->show & SHOW_IT) == SHOW_IT ) {
 			mystrncpy( buf, ip->query, MIDLEN);
@@ -311,7 +378,27 @@ char		*insert_ptr;
 }
 
 
+/*
+ * static int isinlist( item, list)
+ * If "item" is in a comma separated list "list" then return TRUE
+ * else return FALSE.
+ */
 
+static int
+isinlist( item, list)
+char	*item,
+	*list;
+{
+	char	c,
+		*cp;
+
+	if ( (cp = strstr( list, item)) == NULL)
+		return FALSE;
+	c = *(cp + strlen( item));
+	if ( ( c == ',') || (c == '\0') )
+		return TRUE;
+	return FALSE;
+}
 
 
 /*
@@ -337,12 +424,12 @@ File_or_Pipe	*frp;
 		name++;
 		exec_ok( ip);
 		if ( getfpath( buf, name, ip) == FALSE) {
-			logerr( ERRMSG36, name);
+			logerr( err_m[36], name);
 			return NULL;
 		}
 		check_perm( ip, buf);
 		if ( (ofp = popen( buf, "r")) == (FILE *) NULL ) {
-			logerr( ERRMSG36, name);
+			logerr( err_m[36], name);
 			return NULL;
 		}
 		frp->fp = ofp;
@@ -351,12 +438,12 @@ File_or_Pipe	*frp;
 	}
 		
 	if ( getfpath( buf, name, ip) == FALSE) {
-		logerr( ERRMSG1, buf);
+		logerr( err_m[1], buf);
 		return NULL;
 	}
 	check_perm( ip, buf);
 	if ( (ofp = fopen( buf, "r")) == (FILE *) NULL ) {
-		logerr( ERRMSG1, buf);
+		logerr( err_m[1], buf);
 		return NULL;
 	}
 	frp->fp = ofp;
@@ -367,12 +454,11 @@ File_or_Pipe	*frp;
 /*
  * void do_nomatchsub( ip, location)
  * If the return is empty (e.g. a search with no matches) substitute 
- * another local file or a  URL specified in location.  If location
- * begins with '/' it is assumed to be a relative URL and sendredirect is
- * called.  This is necessary even if it is local since otherwise
- * the client would be confused about relative URLs.  If location
+ * another local file.   If location
  * has no '/' then assume it is a file in the same directory and
- * process it as a URL immediately.
+ * process it as a URL immediately.  Error otherwise.  We can't
+ * use a redirect because header (http_prolog) has already been
+ * sent (via search_prolog) before we get here.
  */
 
 void
@@ -383,34 +469,35 @@ char	*location;
 	Parsedata	pdata;
 	File_or_Pipe	frp_st;
 
-
-	if ( *location == '/') {
-		sendredirect( ip, "302 Moved Temporarily", location);
-		return;  /* to send_nomatchsub */
+	if ( strchr( location, '/') != NULL ) {
+		logerr( err_m[108], location);
+		return;
 	}
+
+	if ( openfp( location, ip, &frp_st) == (FILE *) NULL ) {
+		logerr( err_m[83], location);
+		return;
+	}
+
 	pdata.currline = 0;
 	pdata.currfile = location;
 	pdata.show = SHOW_IT;
 
-	if ( openfp( location, ip, &frp_st) == (FILE *) NULL ) {
-		logerr( ERRMSG83, location);
-		return;
-	}
-
 	ip->fp = frp_st.fp;
 	ip->fptype = frp_st.type;
 
+	ip->filetype |= WN_ISHTML;
 	ip->attributes |= WN_PARSE;
 	send_fp( ip, &pdata);
 
-	writelog( ip, LOGMSG2, location);
+	writelog( ip, log_m[2], location);
 	return;		/* to send_nomatchsub */
 }
 
 
 
 /*
- * send_markline_doc( ip, show) Send the main file main file when 
+ * send_markline_doc( ip, show) Send the main file when 
  * ip->type = RTYPE_MARKLINE.
  *
  */
@@ -436,8 +523,8 @@ unsigned	show;
 		insert_param[SMALLLEN],
 		*astart,
 		*s3,
-		*begin,
-		*end;
+		*begin = NULL,
+		*end = NULL;
 
 	int	token;
 
@@ -463,7 +550,7 @@ unsigned	show;
 	off2 = (unsigned) atol( cp);
 
 	if ( off1 > off2) {
-			senderr( CLIENT_ERR, ERRMSG38, "");
+			senderr( CLIENT_ERR, err_m[38], "");
 			wn_exit( 2);
 	}
 
@@ -499,13 +586,13 @@ unsigned	show;
 				end = s3 + off2;
 			}
 			else {
-				senderr( CLIENT_ERR, ERRMSG38, "");
+				senderr( CLIENT_ERR, err_m[38], "");
 				wn_exit( 2);
 			}
 
 
 
-			if ( astart = in_anchor( bigbuf, begin)) {
+			if ( (astart = in_anchor( bigbuf, begin))) {
 				*astart = '\0';
 				send_text_line( bigbuf);
 				sprintf( linebuf, "<a name=\"%s\">&#160;</a>",
@@ -541,7 +628,7 @@ char	*buf,
 {
 	register char	*cp;
 
-	char	*start;
+	char	*start = NULL;
 	int	in = FALSE;
 
 	cp = buf;
@@ -572,6 +659,12 @@ char	*buf,
  * "if", "else", "endif", "start" or "end".  If it does adjust
  * the variable "show" appropriately and return TRUE.  Else return
  * FALSE. 
+ * The "if-elif-else-endif" constructs are parsed on the char
+ * array stack[].  The lower four bits of a value represent the
+ * language token (PS_IF, PS_ELIF, etc) and are obtained by masking
+ * with PARSE_MASK.  The upper four bits (mask with STATE_MASK)
+ * represent the current state, e.g. SS_SHOW = showing text,
+ * SS_HIDDEN = not showing because of enclosing if construct, etc.
  */
 
 int
@@ -579,100 +672,122 @@ set_show( token, pdp)
 int		token;
 Parsedata	*pdp;
 {
+	static unsigned char	stack[STACKSIZE] = { SS_SHOW };
+	unsigned char		parse;
 
-	if ( pdp->show & SHOW_MSK) {
-		switch ( token) {
-		case IF_TRUE:
-			ifcnt++;
-			break;
+	if ( scnt >= STACKSIZE -1 ) {
+		parse_html_err( PARSE_TOO_DEEP, pdp);
+		scnt = 0;
+	}
 
-		case IF_ERR:
-			pdp->show |= ERR_MSK;  /* fall through */
-		case IF_FALSE:
+	switch ( token) {
+	case IF_ERR:
+		pdp->show |= ERR_MSK;
+		pdp->show &= ~SHOW_MSK;
+		break;
+
+	case IF_TRUE:
+	case IF_FALSE:
+		scnt++;
+		stack[scnt] = PS_IF;
+
+		switch (STATE_MASK & stack[scnt-1]) {
+
+		case SS_HIDDEN:
+		case SS_NOSHOW:
+		case SS_SEEN:
+			stack[scnt] |= SS_HIDDEN;
 			pdp->show &= ~SHOW_MSK;
-			ifcnt++;
 			break;
 
-		case ELSE:
-			if ( ifcnt > 0 ) {
+		case SS_SHOW:
+			if ( token == IF_TRUE) {
+				stack[scnt] |= SS_SHOW;
+				pdp->show |= SHOW_MSK;
+			}
+			else {
+				stack[scnt] |= SS_NOSHOW;
 				pdp->show &= ~SHOW_MSK;
 			}
-			else
-				parse_html_err( token, pdp);
 			break;
-
-		case ENDIF:
-			if ( ifcnt > 0 )
-				ifcnt--;
-			else
-				parse_html_err( token, pdp);
-			break;
-
-		case START:
-			if ( !(pdp->show & SECT_MSK))
-				break;
-			pdp->show |= IN_SECT_MSK;
-			break;
-
-		case END:
-			if ( !(pdp->show & SECT_MSK))
-				break;
-			pdp->show &= ~IN_SECT_MSK;
-			break;
-
-		default:
-			return (FALSE);
 		}
-	}
-	else {  			/* SHOW_MSK is off */
-		switch ( token) {
-		case IF_ERR:
-			pdp->show |= ERR_MSK;  /* fall through */
-		case IF_TRUE:
-		case IF_FALSE:
-			ifcnt++;
-			hidden_ifcnt++;
+		break;
+
+	case ELSE:
+	case ELIF_TRUE:
+	case ELIF_FALSE:
+		parse = (PARSE_MASK & stack[scnt]);
+		if ( (parse != PS_IF) && (parse != PS_ELIF)) {
+			parse_html_err( token, pdp);
 			break;
-
-		case ELSE:
-			if ( ifcnt <= 0 )
-				parse_html_err( token, pdp);
-			
-			if ( (hidden_ifcnt <= 0) && !(pdp->show & ERR_MSK) )
-				pdp->show |= SHOW_MSK;
-			break;
-
-		case ENDIF:
-			if ( ifcnt > 0 )
-				ifcnt--;
-			else
-				parse_html_err( token, pdp);
-
-			if ( hidden_ifcnt > 0 )
-				hidden_ifcnt--;
-			else {
-				pdp->show &= ~ERR_MSK;
-				pdp->show |= SHOW_MSK;
-			}
-			break;
-
-		case START:
-			if ( !(pdp->show & SECT_MSK))
-				break;
-			pdp->show |= IN_SECT_MSK;
-			break;
-
-		case END:
-			if ( !(pdp->show & SECT_MSK))
-				break;
-			pdp->show &= ~IN_SECT_MSK;
-			break;
-
-		default:
-			return (FALSE);
 		}
+		scnt++;
+		stack[scnt] = (token == ELSE ? PS_ELSE : PS_ELIF);
+
+		switch (STATE_MASK & stack[scnt-1]) {
+
+		case SS_HIDDEN:
+			stack[scnt] |= SS_HIDDEN;
+			break;
+
+		case SS_SHOW:
+		case SS_SEEN:
+			stack[scnt] |= SS_SEEN;
+			break;
+
+		case SS_NOSHOW:
+			if ( (token == ELIF_TRUE ) || (token == ELSE ))
+				stack[scnt] |= SS_SHOW;
+			else
+				stack[scnt] |= SS_NOSHOW;
+		}
+
+		if ( (stack[scnt] & STATE_MASK) == SS_SHOW )
+			pdp->show |= SHOW_MSK;
+		else
+			pdp->show &= ~SHOW_MSK;
+		break;
+
+	case ENDIF:
+		if ( (stack[scnt] & PARSE_MASK) == PS_ELSE )
+			scnt--;
+
+		while ( (stack[scnt] & PARSE_MASK) == PS_ELIF ) 
+			scnt--;
+
+		if ( (stack[scnt] & PARSE_MASK) == PS_IF ) {
+			scnt--;
+			if ( (scnt < 0) ||
+					(stack[scnt] & STATE_MASK) == SS_SHOW )
+				pdp->show |= SHOW_MSK;
+			else
+				pdp->show &= ~SHOW_MSK;
+
+			break;
+		}
+		else
+			parse_html_err( token, pdp);
+
+		break;
+
+
+	case START:
+		if ( !(pdp->show & SECT_MSK))
+			break;
+		pdp->show |= IN_SECT_MSK;
+		break;
+
+	case END:
+		if ( !(pdp->show & SECT_MSK))
+			break;
+		pdp->show &= ~IN_SECT_MSK;
+		break;
+
+	default:
+		return (FALSE);
 	}
 
+
 	return (TRUE);
 }
 
@@ -689,20 +804,25 @@ Parsedata	*pdp;
 
 	switch ( token) {
 		case ELSE:
-			logerr( PARSE_ERRMSG1, buf);
+		case ELIF_TRUE:
+		case ELIF_FALSE:
+			logerr( parserr_m[1], buf);
 			break;
 		case ENDIF:
-			logerr( PARSE_ERRMSG2, buf);
+			logerr( parserr_m[2], buf);
 			break;
 
 		case START:
-			logerr( PARSE_ERRMSG3, buf);
+			logerr( parserr_m[3], buf);
 			break;
 		case END:
-			logerr( PARSE_ERRMSG4, buf);
+			logerr( parserr_m[4], buf);
+			break;
+		case PARSE_TOO_DEEP:
+			logerr( parserr_m[12], buf);
 			break;
 		case NULLTOKEN:
-			logerr( PARSE_ERRMSG5, buf);
+			logerr( parserr_m[5], buf);
 			break;
 	}
 }
@@ -727,9 +847,11 @@ Parsedata	*pdp;
 	sgmlprocess = (strncasecmp( cp, "<?wn", 4) == 0);
 	oldprocess = (strncmp( cp, "<!--", 4) == 0 );
 
-	if ( !( sgmlprocess || oldprocess) ) {
+	if ( !( (this_rp->filetype & WN_ISHTML) 
+				&& (sgmlprocess || oldprocess) )) {
 		if ( (pdp->show & SHOW_IT) == SHOW_IT) {
-			http_prolog();
+			if ( !(this_rp->status & WN_PROLOGSENT) )
+				http_prolog();
 		}
 		return (NULLTOKEN);
 	}
@@ -741,7 +863,8 @@ Parsedata	*pdp;
 	if ( oldprocess) {
 		if ( *cp != '#' ) {
 			if ( (pdp->show & SHOW_IT) == SHOW_IT) {
-				http_prolog();
+				if ( !(this_rp->status & WN_PROLOGSENT) )
+					http_prolog();
 			}
 			return (COMMENT);
 		}
@@ -794,28 +917,71 @@ Parsedata	*pdp;
 		return (eval_if( cp + 3, pdp));
 
 	}
-	if ( strncasecmp( cp, "include", 7) == 0 ) {
-		return (INCLUDE);
+
+	if ( strncasecmp( cp, "elif ", 5) == 0 ) {
+		int	val;
+
+		val = eval_if( cp + 5, pdp);
+		switch (val) {
+		case IF_TRUE:
+			return	ELIF_TRUE;
+		case IF_FALSE:
+			return	ELIF_FALSE;
+		default:
+			return val;
+		}
 	}
+
+	if ( strncasecmp( cp, "include", 7) == 0 ) {
+		if (this_rp->attrib2 & (WN_WRAPPED + WN_INCLUDE + WN_ISSEARCH))
+			return (INCLUDE);
+
+		cp += 7;
+
+		while ( isspace( *cp) || (*cp == '=') || (*cp == '"'))
+			cp++;
+		mystrncpy( param_ptr, cp, SMALLLEN);
+		cp = param_ptr;
+		while ( *cp && !isspace( *cp) && (*cp != '"'))
+			cp++;
+		*cp = '\0';
+		return (LIST_INCLUDE);
+	}
+
+
 	if ( strncasecmp( cp, "query", 5) == 0 ) {
 		return (QUERY);
 	}
 	if ( strncasecmp( cp, "redirect", 8) == 0 ) {
 		cp += 8;
+		if ( (pdp->show & SHOW_IT) != SHOW_IT)
+			return (REDIRECT);
+
 		if ( (cp = strchr( cp, '"')) == NULL) {
-			logerr( PARSE_ERRMSG11, s);
+			logerr( parserr_m[11], s);
 			return (COMMENT);
 		}
 		mystrncpy( outheadp->redirect, ++cp, MIDLEN);
 		if ( (cp = strchr( outheadp->redirect, '"')) == NULL) {
-			logerr( PARSE_ERRMSG11, s);
+			logerr( parserr_m[11], s);
 			return (COMMENT);
 		}
 		*cp = '\0';
 		return (REDIRECT);
 	}
 	if ( strncasecmp( cp, "section", 7) == 0 ) {
-		return (SECTION);
+		if (this_rp->attrib2 & (WN_WRAPPED + WN_INCLUDE + WN_ISSEARCH))
+			return (SECTION);
+
+		cp += 7;
+		while ( isspace( *cp) || (*cp == '=') || (*cp == '"'))
+			cp++;
+		mystrncpy( param_ptr, cp, SMALLLEN);
+		cp = param_ptr;
+		while ( *cp && !isspace( *cp) && (*cp != '"'))
+			cp++;
+		*cp = '\0';
+		return (LIST_SECTION);
 	}
 	if ( strncasecmp( cp, "search_on", 9) == 0 ) {
 		return (SEARCH_ON);
@@ -832,7 +998,7 @@ Parsedata	*pdp;
 
 	mystrncpy( buf, s, SMALLLEN);
 	chop( buf);
-	logerr( PARSE_ERRMSG5, buf);
+	logerr( parserr_m[5], buf);
 	return (NULLTOKEN);
 }
 
@@ -861,24 +1027,29 @@ Request	*ip;
 {
 	char	linebuf[BIGLEN],
 		insert_param[SMALLLEN],
-		*wrapfile;
+		*wrapfile = NULL;
 	int	token;
 
 	Parsedata	*pdp;
-	static int	firsttime = TRUE;
+	static int	sw_status;
 
 	pdp = &swpdata;
 
+	if (sw_status == SWSTAT_ERROR)
+		return;
+
+	ip->filetype |= WN_ISHTML;
 	if ( swrapfp_st.fp == NULL) {  /* It's the first time so initialize */
 		pdp->currline = 0;
 		pdp->show = SHOW_IT;
-		wrapfile = ( (iswndir( ip)) ? dir_p->swrapper :
-						ip->swrapper);
+		sw_status = SWSTAT_NO_INCL;
+		scnt = 0;
+		wrapfile = ( (iswndir( ip)) ? dir_p->swrapper :	ip->swrapper);
 		pdp->currfile = wrapfile;
 
 		if ( openfp( wrapfile, ip, &swrapfp_st) == NULL ) {
-			senderr( SERV_ERR, ERRMSG46, wrapfile);
-			wn_exit( 2);
+			senderr( SERV_ERR, err_m[46], wrapfile);
+			return;
 		}
 	}
 
@@ -886,13 +1057,12 @@ Request	*ip;
 		pdp->currline++;
 		token = get_parse_token( linebuf, insert_param, pdp);
 		if ( token == INCLUDE) {
-			if (firsttime ) {
-				firsttime = FALSE;
+			if (sw_status == SWSTAT_NO_INCL ) {
+				sw_status = SWSTAT_INCL;
 				return;
 			}
 			else {
-				logerr( ERRMSG84, wrapfile);
-				wn_exit( 2);
+				logerr( err_m[84], wrapfile);
 			}
 		}
 
@@ -912,6 +1082,15 @@ Request	*ip;
 		pclose( swrapfp_st.fp);
 	else
 		fclose( swrapfp_st.fp);
+
+	if (sw_status == SWSTAT_NO_INCL) {
+		sw_status = SWSTAT_ERROR;
+		logerr( err_m[109], wrapfile);
+	}
+
+	/* re-initialize for persistent connections */
+	swrapfp_st.fp = NULL;
+	swrapfp_st.type = FRP_FILE;
 }
 
 
@@ -931,13 +1110,13 @@ Request	*ip;
 	static int	num = 0;
 
 	char	*cp,
-		buf[2*MIDLEN],
-		loc[2*MIDLEN];
+		*lhost,
+		buf[2*MIDLEN];
 
 	num++;
 
 	if ( num > MAX_REDIRECT) {
-		senderr( SERV_ERR, ERRMSG55, "");
+		senderr( SERV_ERR, err_m[55], "");
 		wn_exit(2);
 	}
 
@@ -946,6 +1125,8 @@ Request	*ip;
 		return;
 	}
 
+	lhost = ( *(inheadp->host_head) ? (inheadp->host_head) : hostname );
+
 	mystrncpy( buf, path, MIDLEN);
 
 	if ( *(ip->query) 
@@ -954,47 +1135,49 @@ Request	*ip;
 	/* ends with '?' so add the ip->query to it */
 
 	if ( ((cp = strchr( buf, ':')) != NULL) && (*(cp+1) == '/'))
-		{ /* URL rediretion */
-		sendredirect( ip, "302 CGI redirection", buf);
+		{ /* URL rediretion: it has "http://" or equivalent  */
+		sendredirect( ip, "302 Redirection", buf);
 		return;
 	}
 
-	if ( *buf == '/') {
-		if ( port == STANDARD_PORT )
-			sprintf( loc, "%s://%s", this_conp->scheme, hostname);
-		else
-			sprintf( loc, "%s://%s:%d", this_conp->scheme, 
-					hostname, port);
-
-		mystrncat( loc, buf, MIDLEN);
-		sendredirect( ip, "302 CGI redirection", loc);
-		return;
-	}
-
-
-	mystrncpy( loc, inheadp->url_path, MIDLEN);
-	cp = strrchr( loc, '/');
-	mystrncpy( ++cp, buf, MIDLEN);
-
-	if ( (strchr( path, '#') != NULL) || (strchr( path, '/') != NULL)) {
+	if ( (strchr( buf, '#') != NULL) || (strchr( buf, '/') != NULL)) {
 		/* Relative URL with anchor designator requires redirect */
 		/* and so does any relative URL with a '/'  */
-		if ( port == STANDARD_PORT )
-			sprintf( loc, "%s://%s", this_conp->scheme, hostname);
+
+		char	loc[BIGLEN];   /* BIGLEN = 2*SMALLLEN */
+
+		if ( (port == STANDARD_PORT) ||
+					( strchr( lhost, ':') != NULL) )
+			sprintf( loc, "%s://%.100s", this_conp->scheme, lhost);
 		else
-			sprintf( loc, "%s://%s:%d", this_conp->scheme, 
-					hostname, port);
+			sprintf( loc, "%s://%.100s:%d", this_conp->scheme, 
+					lhost, port);
 
-		mystrncat( buf, loc, MIDLEN);
-		sendredirect( ip, "302 CGI redirection", buf);
+		if ( *buf == '/') {
+			mystrncat( loc, buf, MIDLEN);
+			sendredirect( ip, "302 CGI redirection", loc);
+			return;
+		}
+
+		mystrncat( loc, inheadp->url_path, MIDLEN);
+		cp = strrchr( loc, '/');
+		mystrncpy( ++cp, buf, MIDLEN - SMALLLEN);
+		sendredirect( ip, "302 Redirection", loc);
 		return;
+
 	}
+	else {
+		char	sloc[BIGLEN];
 
+		mystrncpy( sloc, inheadp->url_path, MIDLEN);
+		cp = strrchr( sloc, '/');
+		mystrncpy( ++cp, buf, MIDLEN);
 
-	*(outheadp->redirect) = '\0';
-	ip->do_wrap_1st_time = TRUE;
+		*(outheadp->redirect) = '\0';
+		ip->do_wrap_1st_time = TRUE;
 
-	process_url( ip, loc);
+		process_url( ip, sloc);
+	}
 
 	if ( ip->attributes & (WN_PARSE + WN_DYNAMIC + WN_FILTERED) ) {
 		wn_exit(0);
