@@ -1,7 +1,7 @@
 /*
     Wn: A Server for the HTTP
     File: wn/prequest.c
-    Version 1.17.0
+    Version 1.18.3
 
     Copyright (C) 1996  <by John Franks>
 
@@ -76,6 +76,12 @@ char	*url_path;
 	char		redirect[MIDLEN],
 			path[MIDLEN];
 
+	if ( (inheadp->protocol ==  HTTP1_1) &&	(!*inheadp->host_head) ) {
+		/* Missing Host: header in 1.1 request is an error */
+		senderr(CLIENT_ERR, err_m[15], "");
+		return;
+	}
+
 	ip->type = RTYPE_UNCHECKED; /* we don't know type yet */
 
 	/*
@@ -155,6 +161,9 @@ char	*url_path;
 
 	path_security( ip, path, FALSE);
 
+	if ( ip->type == RTYPE_FINISHED)
+		return;
+
 	cp = path;
 	/* move cp to end of path */
 	while ( *cp)
@@ -170,13 +179,12 @@ char	*url_path;
 
 	/* Fill in filepath, and pointers relpath and basename */
 
-	strcpy( ip->filepath, ip->rootdir);
+	mystrncpy( ip->filepath, ip->rootdir, MAXDIRLEN);
 	cp = ip->filepath;
 	while ( *cp)
 		cp++;
 	ip->relpath = cp;
-	mystrncpy( ip->relpath, path, MIDLEN - MAXDIRLEN);
-			/* path has size BIGLEN! */
+	mystrncpy( ip->relpath, path, MIDLEN - (MAXDIRLEN + TINYLEN + 4));
 
 	cp = strrchr( ip->filepath, '/');
 	ip->basename = ++cp;
@@ -191,22 +199,25 @@ char	*url_path;
 			while ( *cp && isspace( *cp))
 				cp++;
 			if ( strncasecmp( inheadp->range, "bytes", 5) != 0 ) {
-				senderr(CLIENT_ERR, ERRMSG103, inheadp->range);
-				wn_exit( 2);
+				senderr(CLIENT_ERR, err_m[103],
+							inheadp->range);
+				return;
 			}
 			mystrncpy( ip->range, cp, RANGELEN);
 			ip->filetype |= (WN_BYTERANGE + WN_RFC_BYTERANGE);
 		}
 		else {
-			senderr( CLIENT_ERR, ERRMSG95, inheadp->range);
-			wn_exit( 2);
+			senderr( CLIENT_ERR, err_m[95], inheadp->range);
+			return;
 		}
 	}
 
 	set_param( ip);
 
+	if ( ip->type == RTYPE_FINISHED)
+		return;
 	if (ip->type == RTYPE_DENIED) {
-		logerr( ERRMSG85, ip->param_field);
+		logerr( err_m[85], ip->param_field);
 		return;
 	}
 
@@ -221,21 +232,29 @@ char	*url_path;
 		switch ( ip->type) {
 		case RTYPE_CGI:
 		case RTYPE_NPH_CGI:
-				senderr( SERV_ERR, ERRMSG102, url_path);
-				wn_exit( 2);
+			senderr( SERV_ERR, err_m[102], url_path);
+			return;
+		default:
+			break;
 		}
 		if ( !trailslash) {
+			char	*lhost;
+
+			lhost = ( *(inheadp->host_head) ?
+				(inheadp->host_head) : hostname );
+
 			if ( strchr( url_path, ';') || 	strchr( url_path, '?')
 					 || strchr( url_path, '=') ) {
-				senderr( CLIENT_ERR, ERRMSG59, path);
-				wn_exit( 2);
+				senderr( CLIENT_ERR, err_m[59], path);
+				return;
 			}
-			if ( port == STANDARD_PORT )
+			if ( (port == STANDARD_PORT) ||
+					( strchr( lhost, ':') != NULL) )
 				sprintf( redirect, "%s://%s%s/", 
-					this_conp->scheme, hostname, url_path);
+					this_conp->scheme, lhost, url_path);
 			else
 				sprintf( redirect, "%s://%s:%d%s/",
-						this_conp->scheme, hostname,
+						this_conp->scheme, lhost,
 						port, url_path);
 			sendredirect( ip, "301 Moved Permanently", redirect);
 			ip->type = RTYPE_FINISHED;
@@ -246,11 +265,14 @@ char	*url_path;
 		case RTYPE_INFO:
 			cp = ip->relpath;
 			strcat ( ip->relpath, "/");
-			strcat ( ip->relpath, INDEXFILE_NAME);
+			mystrncat ( ip->relpath, INDEXFILE_NAME, TINYLEN);
 			ip->filetype &= ~(WN_DIR);
 			ip->filetype |= WN_DEFAULT_DOC;
 			get_stat( ip);
 			ip->basename = strrchr( ip->relpath, '/') + 1;
+			break;
+		default:
+			break;
 		}
 	}
 	else if ( trailslash) {
@@ -332,8 +354,8 @@ Request	*ip;
 			if ( strncasecmp( ip->param_value, "field", 5) == 0)
 				ip->type = RTYPE_FIELDSEARCH;
 			else {
-				senderr( SERV_ERR, ERRMSG58, ip->param_value);
-				wn_exit( 2);
+				senderr( SERV_ERR, err_m[58], ip->param_value);
+				return;
 			}
 			break;
 		case 'g':
@@ -371,7 +393,8 @@ Request	*ip;
  * Stats the file pointed to by ip->filepath.  If it fails assume file
  * does not exist.  If it is a directory set WN_DIR bit in ip->filetype
  * unless this is the default document, in which case log an error and
- * deny access. If not a directory get length and modification time.
+ * deny access. If not a directory make sure it is a regular file and
+ * if so, get length and modification time and set the etag.
  */
 
 void
@@ -380,7 +403,11 @@ Request	*ip;
 {
 	struct stat stat_buf;
 
-	if ( stat( ip->filepath, &stat_buf) != 0 ) {
+
+#ifndef S_ISREG
+#define	S_ISREG(m)	(((m)&S_IFMT) == S_IFREG)
+#endif
+	if ( lstat( ip->filepath, &stat_buf) != 0 ) {
 		/* Might be redirect, or defdoc, just continue */
 		*ip->length = '\0';
 		ip->mod_time = 0;
@@ -388,21 +415,49 @@ Request	*ip;
 		return;
 	}
 
+	if ( (!S_ISREG(stat_buf.st_mode)) && (!S_ISDIR(stat_buf.st_mode))) { 
+		/* probably a symlink */
+		if ( stat( ip->filepath, &stat_buf) != 0 ) {
+			*ip->length = '\0';
+			ip->mod_time = 0;
+			ip->status |= WN_CANT_STAT;
+			return;
+		}
+		ip->filetype |= WN_NOT_REG_FILE;
+	}
+
 	if ( !(stat_buf.st_mode & S_IROTH))
-		ip->status |= WN_NOT_WORLD_READ;
+		ip->filetype |= WN_NOT_WORLD_READ;
+
 
 	if ( S_ISDIR( stat_buf.st_mode)) {
 		if ( ip->filetype & WN_DEFAULT_DOC) {
-			logerr( ERRMSG88, ip->filepath);
+			logerr( err_m[88], ip->filepath);
 			ip->type = RTYPE_DENIED;
 		}
 		else
 			ip->filetype |= WN_DIR;
 		return;
 	}
+
+	if ( !(S_ISREG(stat_buf.st_mode))) {
+		logerr( err_m[115], ip->filepath);
+		ip->type = RTYPE_DENIED;
+		return;
+	}
+
 	ip->mod_time = stat_buf.st_mtime;
-	ip->datalen = stat_buf.st_size;
-	sprintf( ip->length, "%lu", (unsigned long) stat_buf.st_size);
+	ip->datalen = (long) stat_buf.st_size;
+	sprintf( ip->length, "%lu", ip->datalen);
+	set_etag( &stat_buf);
+}
+
+void
+set_etag( sbp)
+struct stat *sbp;
+{
+	sprintf( this_rp->etag, "%lx-%lx-%lx", (long) sbp->st_mtime,
+				(long) sbp->st_ino, (long) sbp->st_size);
 }
 
 static void
@@ -412,6 +467,9 @@ char	*path;
 {
 	register char	*cp;
 	
+	if ( ip->type == RTYPE_FINISHED)
+		return;
+
 	*(ip->pathinfo) = '\0';
 	if ( (cp = strstr( path, CGI_EXT)) != NULL) {
 		cp += (sizeof( CGI_EXT) - 1);
@@ -427,7 +485,7 @@ char	*path;
 #ifdef CGI_BIN
 	if ( (cp = strstr( path, CGI_BIN)) != NULL) {
 		cp--;
-		if ( *cp != '/')
+		if ( (*cp != '/') || ( cp < path))
 			return;
 		cp += sizeof( CGI_BIN);
 		if ( *cp == '/') {
@@ -451,11 +509,14 @@ char	*path;
 {
 	register char	*cp;
 
+	if ( ip->type == RTYPE_FINISHED)
+		return;
+
 	/* First handle ';' delimiter (for files) */
 	if ( ( cp = strrchr( path, ';')) != NULL) {
 		if ( *(cp - 1) == '/' ) {  /* path with /; is error */
-			senderr(CLIENT_ERR, ERRMSG59, path);
-			wn_exit( 2);
+			senderr(CLIENT_ERR, err_m[59], path);
+			return;
 		}
 		*cp++  = '\0';
 	}
@@ -494,6 +555,7 @@ char	*path;
  * dedot() has already handled "/../" and "/./".  We also check that
  * path is either empty (referring to root) or begins with '/'.
  * Things are a little more lax for the PATHINFO part of a CGI URL.
+ * At the moment this means that '=' is allowed.
  */
 
 static void
@@ -508,39 +570,40 @@ int	is_pathinfo;
 	/* Security check */
 	cp = path;
 	if ( *cp && (*cp != '/') ) {
-		senderr( CLIENT_ERR, ERRMSG59, path);
-		wn_exit( 2);
+		senderr( CLIENT_ERR, err_m[59], path);
+		return;
 	}
 
 	while ( *cp ) {
-		switch( *cp) {
-		case '.':
-			if ( (cp[1] == '.') && (cp[2] == '/')) {
-				logerr( ERRMSG21, ip->request);
-				senderr( CLIENT_ERR, ERRMSG59, path);
-				wn_exit( 2);
-			}
-		case '/':
-		case '-':
-		case '+':
-		case '_':
-		case '%':
+		if ( myisalnum( *cp))
 			cp++;
-			break;
-		/* Stuff allowed for pathinfo here */
-		case '=':
-			if ( is_pathinfo) {
+		else {
+			switch( *cp) {
+			case '.':
+				if ( (cp[1] == '.') && (cp[2] == '/')) {
+					logerr( err_m[21], ip->request);
+					senderr( CLIENT_ERR, err_m[59], path);
+					return;
+				}
+			case '/':
+			case '-':
+			case '+':
+			case '_':
+			case '%':
 				cp++;
 				break;
-			}
-		default:
-			if ( !isalnum( *cp)) {			
-				sprintf( buf, ERRMSG20,	*cp);
+			/* Stuff allowed for pathinfo here */
+			case '=':
+				if ( is_pathinfo) {
+					cp++;
+					break;
+				}
+			default:
+				sprintf( buf, err_m[20],	*cp);
 				logerr( buf, ip->request);
-				senderr( CLIENT_ERR, ERRMSG59, path);
-				wn_exit( 2);
+				senderr( CLIENT_ERR, err_m[59], path);
+				return;
 			}
-			cp++;
 		}
 	}
 }
@@ -549,7 +612,7 @@ int	is_pathinfo;
 /*
  * dedot( ip, path) replaces the first "//", "/./", and "/xxxx/../" 
  * in path with "/".  It returns TRUE if it found one of these
- * and FALSE if there weren none of these in path.
+ * and FALSE if there were none of these in path.
  */
 
 static int
@@ -560,6 +623,9 @@ char	*path;
 	register char	*cp,
 			*cp2;
 
+	if ( ip->type == RTYPE_FINISHED)
+		return FALSE;
+
 	cp = path;
 	while ( (cp = strchr( cp, '/')) != NULL) {
 		cp++;
@@ -574,10 +640,10 @@ char	*path;
 		if ( strncmp( cp, "../", 3) == 0 ) {
 			*--cp = '\0';
 			if ( (cp2 = strrchr( path, '/')) == NULL) {
-				logerr( ERRMSG21, ip->request);
-				senderr( CLIENT_ERR, ERRMSG59, path);
-				wn_exit( 2);
-			}
+				logerr( err_m[21], ip->request);
+				senderr( CLIENT_ERR, err_m[59], path);
+				return FALSE;
+		}
 			strcpy( cp2, cp + 3);
 			return TRUE;
 		}
@@ -602,7 +668,7 @@ Request	*ip;
 {
 
 #ifdef LIMIT_2_HIERARCHY
-	return getfpath2( path, fname, ip);
+	return getfpath2( path, fname, ip->cachepath);
 #else
 	register char	*cp;
 
@@ -616,7 +682,7 @@ Request	*ip;
 	else {
 		strcpy( path, ip->cachepath);
 		if ( (cp = strrchr( path, '/')) == NULL) {
-			logerr( ERRMSG37, path);
+			logerr( err_m[37], path);
 			return FALSE;
 		}
 		else 
@@ -628,17 +694,17 @@ Request	*ip;
 
 
 /*
- * getfpath2( path, fname, ip) is like getfpath except it restricts
+ * getfpath2( path, fname, cachepath) is like getfpath except it restricts
  * to WN hierarchy.  If fname starts with either '/' or '~/' then
  * it is assumed relative to WN root  otherwise it is assumed 
  * relative to current directory.
  */
 
 int
-getfpath2( path, fname, ip)
+getfpath2( path, fname, cachepath)
 char	*path,
-	*fname;
-Request	*ip;
+	*fname,
+	*cachepath;
 {
 
 	register char	*cp;
@@ -646,21 +712,21 @@ Request	*ip;
 	if ( *fname == '~')
 		fname++;
 	if ( *fname == '/') {
-		mystrncpy( path, ip->rootdir, SMALLLEN);
+		mystrncpy( path, this_rp->rootdir, SMALLLEN);
 		mystrncat( path, fname, MIDLEN - SMALLLEN);
 	}
 	else {
-		strcpy( path, ip->cachepath);
+		strcpy( path, cachepath);
 		if ( (cp = strrchr( path, '/')) == NULL) {
-			logerr( ERRMSG37, path);
+			logerr( err_m[37], path);
 			return FALSE;
 		}
 		else 
 			strcpy( ++cp, fname);
 	}				
 	if ( strstr( "../", path) != NULL) {
-		logerr( ERRMSG21, ip->request);
-		wn_exit( 2);
+		logerr( err_m[21], path);
+		return FALSE;
 	}
 	return TRUE;
 }
@@ -723,20 +789,27 @@ Request *ip;
 
 	max_mtime = 0;
 
-	if ( ip->attributes & WN_WRAPPED) {
+	if ( ip->attrib2 & WN_WRAPPED) {
 		mtime = max_mtime_includes(ip->wrappers, ip);
 		if (mtime > max_mtime) {
 			max_mtime = mtime;
 		}
 	}
 
-	if ( ip->attributes & WN_INCLUDE ) {
+	if ( ip->attrib2 & WN_INCLUDE ) {
 		mtime = max_mtime_includes(ip->includes, ip);
 		if (mtime > max_mtime) {
 			max_mtime = mtime;
 		}
 	}
 
+	if ( ip->attrib2 & WN_LIST_INCL ) {
+		mtime = max_mtime_includes(ip->list_incl, ip);
+		if (mtime > max_mtime) {
+			max_mtime = mtime;
+		}
+	}
+
 	if (max_mtime) {
 		/*
 		 * There was a wrapper/include so restat base
@@ -757,12 +830,16 @@ Request *ip;
  * set_interface_root() checks the hostname and IP address to which
  * the client has connected and sets our data root (this_rp->rootdir[]) to the
  * apprpropriate value.  The appropriate value is determined by consulting
- * the array vhostlist[][3] which is defined and initialized in vhost.c.
+ * the array vhostlist[][0] which is defined and initialized in vhost.c.
  * If inheadp->host_head contains something it is a hostname set from
- * either the Host: header or a full URL in the request.  If anything in
- * the vhostlist matches this it is used.  If no hostname has been set
- * by Host: or URL then the IP addresses from vhostlist are searched for
- * a match.  In either case if no match is found the default rootdir
+ * either the Host: header or a full URL in the request.  If any hostname in
+ * the vhostlist matches this then the root dir corresponding to the
+ * first match is used.  If the Host: header or URL contains an IP address
+ * then the first match to vhostlist[][1] is used.  If no hostname has been
+ * set by Host: or URL or if no match as above was found then the IP 
+ * addresses from vhostlist[][1] are searched for a match to the IP address
+ * on which the request was received.  The root dir corresponding to the
+ * first such match is used.  If there is still no match the default rootdir
  * set on the command line or in config.h is used.
  */
 
@@ -779,14 +856,13 @@ set_interface_root( )
 
 	static struct sockaddr_in      saddr;
 
-	ifn = 0;
-
 	if ( *(inheadp->host_head)) {
 		mystrncpy( lhostname, inheadp->host_head, MAXHOSTNAMELEN);
 		if ( (cp = strchr( lhostname, ':')) != NULL) 
 			*cp = '\0';
 		while ( vhostlist[ifn][2] != NULL) {
-			if ( strcasecmp( vhostlist[ifn][0], lhostname) == 0) {
+			if ( (strcasecmp( vhostlist[ifn][0], lhostname) == 0)
+				|| streq( vhostlist[ifn][1] , lhostname))  {
 				mystrncpy( this_rp->rootdir,
 					vhostlist[ifn][2], SMALLLEN);
 				cp = vhostlist[ifn][0];
@@ -796,12 +872,13 @@ set_interface_root( )
 			ifn++;
 		}
 	}
-	else {
+	if ( !*(inheadp->host_head) || (vhostlist[ifn][2] == NULL)) {
+	/* if no host header or no match for host header use IP# */
 		if ( !*local_ip) {
 			size = sizeof(saddr);
 			if ( getsockname( fileno( stdin),
 				(struct sockaddr *) &saddr, &size) < 0 ) {
-				logerr( ERRMSG73, "");
+				logerr( err_m[73], "");
 				mystrncpy(local_ip, "127.0.0.1", 20);
 
 			}
@@ -810,7 +887,7 @@ set_interface_root( )
 					inet_ntoa(saddr.sin_addr), 20);
 			}
 		}
-		while ( vhostlist[ifn][2] != NULL) {
+		for ( ifn = 0; vhostlist[ifn][2] != NULL; ifn++) {
 			if ( streq( vhostlist[ifn][1] , local_ip)) {
 				mystrncpy( this_rp->rootdir,
 					vhostlist[ifn][2], SMALLLEN);
@@ -820,7 +897,6 @@ set_interface_root( )
 				mystrncpy( hostname, cp, SMALLLEN);
 				break;
 			}
-			ifn++;
 		}
 	}
 	interface_num = ((vhostlist[ifn][2] == NULL) ? 0 : ++ifn);
