@@ -1,9 +1,9 @@
 /*
     Wn: A Server for the HTTP
     File: wn/vhost.h
-    Version 1.16.0
+    Version 1.17.3
     
-    Copyright (C) 1996  <by John Franks>
+    Copyright (C) 1996, 1997  <by John Franks>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -22,7 +22,10 @@
 */
 
 #ifdef USE_VIRTUAL_HOSTS
-char *vhostlist[][3] =
+WN_CONST
+char *
+WN_CONST
+vhostlist[][3] =
 {
 	{ "localhost", "127.0.0.1", ROOT_DIR },
 	{ "abc.def.com" , "123.123.121.1", "/usr/local/data1" },
