@@ -1,9 +1,9 @@
 /*
     Wn: A Server for the HTTP
     File: wn/wn.c
-    Version 1.17.0
+    Version 1.18.7
     
-    Copyright (C) 1996  <by John Franks>
+    Copyright (C) 1996-98  <by John Franks>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -35,6 +35,11 @@
 #ifndef NO_UNISTD_H
 #include <unistd.h>
 #endif
+
+#ifndef STANDALONE
+#include <sys/stat.h>
+#endif
+
 #include "wn.h"
 #include "version.h"
 #include "cgi.h"
@@ -43,9 +48,9 @@
 #define WN_read(a,b,c)		read(a,b,c)
 #endif
 
-#define MAX_POST		(2048 * 1024)
+#define MAX_POST		(5 * 1024 * 1024)
 #define KEEP_ALIVE_TIMEOUT	(20)
-#define NETSCAPE2_BUG
+#define BROWSER_BUGS
 #define AUTHENT_TIMEOUT		(30)
 #define KEEP_ALIVE_MAX		(10)
 
@@ -53,10 +58,6 @@
 #define AWAIT_HEADER		(2)
 #define AWAIT_TRANSACTION	(3)
 
-#ifndef STANDALONE
-#include <sys/stat.h>
-#endif
-
 extern int	daemon_init();
 extern long	atol();
 extern void	parse_request(),
@@ -65,8 +66,8 @@ extern void	parse_request(),
 extern time_t	time();
 
 static void	do_post(),
-		reset_buf(),
-		get_header(),
+		get_headers(),
+		chk_method(),
 		wn_timeout(),
 		client_closed();
 
@@ -74,6 +75,7 @@ static Methodtype parse_header( );
 static char	*get_input();
 
 static int	chk_continue(),
+		reset_buf(),
 		load_inbuf(),
 		wn_getc(),
 		await_state,
@@ -126,8 +128,7 @@ do_connection()
 
 	thiscon.keepalive = FALSE;
 	thiscon.trans_cnt = 0;
-	thiscon.authuser[0] = thiscon.rfc931name[0]
-		= thiscon.logbuf[0] = '\0';
+	thiscon.rfc931name[0] = thiscon.logbuf[0] = '\0';
 	thiscon.pid = getpid();
 	readbuf.bcp = NULL;
 	readbuf.cur_sz = 0;
@@ -139,7 +140,7 @@ do_connection()
 	this_conp->log_ptr = this_conp->logbuf;
 	this_conp->out_ptr = this_conp->outbuf;
 	this_conp->scheme = "http";
-	
+	this_conp->con_status = 0;
 
 
 	 /* Start a debug entry if debugging enabled. */
@@ -179,7 +180,7 @@ do_connection()
 
 		await_state = AWAIT_HEADER;
 
-		strcpy( thisreq.request, request);
+		mystrncpy( thisreq.request, request, BIGLEN);
 
 		if ( parse_header( request) == UNKNOWN ) {
 			wn_exit( 0);
@@ -193,10 +194,13 @@ do_connection()
 #ifdef NO_KEEPALIVE
 		this_conp->keepalive = FALSE;
 #else
-#ifdef NETSCAPE2_BUG
-		if ( this_conp->keepalive &&
-			(strncmp( inheadp->ua, "Mozilla/2.0", 11) == 0)) {
-			this_conp->keepalive = FALSE;
+#ifdef BROWSER_BUGS
+		if ( this_conp->keepalive ) {
+			if ( strstr( inheadp->ua, "Mozilla/2.0") ||
+					strstr( inheadp->ua, "MSIE 4.0b") ) {
+				this_conp->keepalive = FALSE;
+				inheadp->protocol =  HTTP1_0;
+			}
 		}
 #endif
 #endif
@@ -207,11 +211,19 @@ do_connection()
 		}
 
 		process_url(&thisreq, inheadp->url_path);
+
+		if ( this_rp->status & WN_ABORTED)
+			wn_exit( 2);
+
 		if ( !(this_conp->keepalive)) {
 			break;
 			/* if not keepalive then exit */
 		}
-		flush_outbuf();
+
+		if ( this_conp->more_in_buf )
+			continue;
+		else
+			flush_outbuf();
 	}
 	wn_exit(0);
 }
@@ -227,10 +239,10 @@ wn_timeout( )
 			 */
 
 	if ( await_state != AWAIT_REQUEST) {
-		strcpy( outheadp->status, "503 Timed Out");
+		strcpy( outheadp->status, "408 Timed Out");
 		sprintf( buf, "Process %d, await_state (%d) ",
 			this_conp->pid, await_state);
-		senderr( "503", ERRMSG60, buf);
+		senderr( "408", err_m[60], buf);
 	}
 	wn_exit( 2);
 }
@@ -242,14 +254,13 @@ client_closed( )
 {
 	if ( debug_log ) {
 		char	buf[SMALLLEN];
-		sprintf( buf, "Client closed connection, pid=%d,  wait=%d ",
-					this_conp->pid, await_state);
+		sprintf( buf, log_m[19], this_conp->pid, await_state);
 		write_debug( 1, buf, "");
 	}
 	if ( await_state != AWAIT_REQUEST) {
 		*(this_rp->length) = '\0';
 		get_remote_info();
-		writelog( this_rp, LOGMSG19, (await_state == AWAIT_HEADER 
+		writelog( this_rp, log_m[18], (await_state == AWAIT_HEADER 
 			? "header" : "transaction"));
 	}
 	wn_exit( 0);
@@ -262,6 +273,7 @@ char	*url_path;
 {
 	Reqtype	itemtype;
 
+
 	/* These may need clearing if we came from redirect */
 	ip->attributes = ip->filetype = ip->status = 0;
 
@@ -273,26 +285,54 @@ char	*url_path;
 	itemtype = ip->type;
 
 	if ( (itemtype != RTYPE_DENIED) && (itemtype != RTYPE_NO_AUTH)
+				&& (itemtype != RTYPE_NOACCESS)
+				&& (itemtype != RTYPE_REDIRECT) 
 				&& (itemtype != RTYPE_FINISHED)) {
 		switch (inheadp->method) {
 		case HEAD:
 			itemtype = RTYPE_HEADER;
 			break;
 		case CONDITIONAL_GET:
-			if ( (ip->attributes & WN_DYNAMIC) 
-				|| date_cmp( ip, inheadp->inmod_date, TRUE))
+			if ( ip->attributes & WN_DYNAMIC) {
 				inheadp->method = GET;
-			else
-				itemtype = RTYPE_NOT_MODIFIED;
-			break;
-		case UNLESS_GET:
+				break;
+			}				
 			inheadp->method = GET;
-			if ( date_cmp( ip, inheadp->inmod_date, TRUE)) {
+
+			if ( (inheadp->conget & IFMODSINCE)
+				&& !date_cmp( ip, inheadp->inmod_date, TRUE))
+				itemtype = RTYPE_NOT_MODIFIED;
+
+			if ( (inheadp->conget & IFUNMODSINCE)
+				&& date_cmp( ip, inheadp->inmod_date, TRUE)) {
+				itemtype = RTYPE_PRECON_FAILED;
+				break;
+			}
+
+			if ( inheadp->conget & IFNMATCH) {
+				if ( streq( inheadp->etag, ip->etag))
+					itemtype = RTYPE_NOT_MODIFIED;
+			}
+
+			if ( inheadp->conget & IFMATCH) {
+				if ( !streq( inheadp->etag, ip->etag))
+					itemtype = RTYPE_PRECON_FAILED;
+			}
+
+			if ( (inheadp->conget & IFRANGE) && *inheadp->etag) {
+				if ( !streq( inheadp->etag, ip->etag))
+					ip->filetype &= ~(WN_RFC_BYTERANGE
+						+ WN_BYTERANGE
+						+ WN_LINERANGE);
+			}
+			else if ( (inheadp->conget & IFRANGE) && 
+				(date_cmp( ip, inheadp->inmod_date, TRUE))) {
 				ip->filetype &= ~(WN_RFC_BYTERANGE
 						+ WN_BYTERANGE
 						+ WN_LINERANGE);
 			}
 			break;
+
 		case POST:
 		case PUT:
 			do_post( inheadp);
@@ -306,6 +346,7 @@ char	*url_path;
 		case RTYPE_FINISHED:
 			break;
 		case RTYPE_FILE:
+			chk_method( GET);
 			if ( ip->attributes & WN_CGI )
 				cgi_env( ip, FALSE);
 
@@ -315,49 +356,61 @@ char	*url_path;
 				sendbin( ip);
 			break;
 		case RTYPE_MARKLINE:
+			chk_method( GET);
 			if ( ip->attributes & WN_CGI )
 				cgi_env( ip, FALSE);
 
 			if ( ip->filetype & WN_TEXT )
 				sendtext( ip);
 			else
-				senderr( DENYSTATUS, ERRMSG47, ip->filepath);
+				senderr( DENYSTATUS, err_m[47], ip->filepath);
 			break;
 #ifndef FORBID_CGI
 		case RTYPE_CGI:
 		case RTYPE_NPH_CGI:
+		case RTYPE_CGI_HANDLER:
 			sendcgi( ip);
 			break;
 #endif
 		case RTYPE_GSEARCH:
 		case RTYPE_CONTEXTSEARCH:
 		case RTYPE_LINESSEARCH:
+			chk_method( GET);
+			ip->attrib2 |= WN_ISSEARCH;
 			sendgrep( ip);
 			break;
 		case RTYPE_TSEARCH:
 		case RTYPE_KSEARCH:
 		case RTYPE_TKSEARCH:
 		case RTYPE_FIELDSEARCH:
+			chk_method( GET);
+			ip->attrib2 |= WN_ISSEARCH;
 			cache_search( ip);
 			break;
 		case RTYPE_ISEARCH:
+			chk_method( GET);
+			ip->attrib2 |= WN_ISSEARCH;
 			if ( ip->attributes & WN_CGI )
 				cgi_env( ip, FALSE);
 			send_isearch( ip);
 			break;
 		case RTYPE_LISTSEARCH:
+			chk_method( GET);
+			ip->attrib2 |= WN_ISSEARCH;
 			list_search( ip);
 			break;
-		case RTYPE_HEADER:
-			http_prolog( );
-			writelog( ip, LOGMSG6, "");
-			break;
 		case RTYPE_IMAGEMAP:
+			chk_method( GET);
 			image( );
 			break;
 		case RTYPE_INFO:
+			chk_method( GET);
 			sendinfo( ip);
 			break;
+		case RTYPE_HEADER:
+			http_prolog( );
+			writelog( ip, log_m[6], "");
+			break;
 		case RTYPE_REDIRECT:
 			sendredirect( ip, "301 Moved Permanently", 
 				outheadp->redirect);
@@ -366,10 +419,15 @@ char	*url_path;
 			strcpy( outheadp->status, "304 Not Modified");
 			http_prolog( );
 			strcpy( ip->length, "0");
-			writelog( ip, LOGMSG5, "");
+			writelog( ip, log_m[5], "");
 			break;
 		case RTYPE_NO_AUTH:
 			break;
+		case RTYPE_PRECON_FAILED:
+			senderr( PRECON_STATUS, out_m[5], ip->filepath);
+			ip->type = RTYPE_FINISHED;
+			break;
+
 		case RTYPE_NOACCESS:
 			if ( *(dir_p->noaccess_url) && 
 				*dir_p->noaccess_url &&
@@ -377,20 +435,44 @@ char	*url_path;
 				sendredirect( ip, "302 Moved Temporarily", 
 					dir_p->noaccess_url);
 				break;
+			/* Falls through */
 			}
 		case RTYPE_DENIED:
-			if ( ip->status & WN_CANT_STAT &&
-				*dir_p->cantstat_url &&
+			if ( 	*dir_p->cantstat_url &&
 				!streq( ip->relpath, dir_p->cantstat_url)) {
 				sendredirect( ip, "302 Moved Temporarily", 
 					dir_p->cantstat_url);
 				break;
+			/* Falls through */
 			}
 		case RTYPE_UNCHECKED:
 		default:
-			senderr( DENYSTATUS, DENYMSG, ip->filepath);
+#ifdef DENYHANDLER
+			{
+				FILE	*fp;
+				cgi_env( ip, FALSE);
+				if ((fp = popen( DENYHANDLER, "r")) != NULL) {
+					ip->attributes |= WN_UNBUFFERED;
+					this_conp->keepalive = FALSE;
+					send_out_fd( fileno( fp));
+					pclose( fp);
+					sprintf( outheadp->status,
+						"%s %s", DENYSTATUS, out_m[0]);
+					writelog( this_rp,
+						out_m[0], ip->filepath);
+					ip->type = RTYPE_FINISHED;
+					ip->status |= WN_ERROR;
+					return;
+				}
+				/*
+				 * If handler cannot be run, fall thru
+				 * to standard error document.
+				 */
+			}
+#endif
+			senderr( DENYSTATUS, out_m[0], ip->filepath);
+			ip->type = RTYPE_FINISHED;
 	}
-
 	return;
 
 	/*
@@ -406,8 +488,7 @@ parse_header( req)
 char	*req;
 {
 	register char	*cp;
-	char		errmsg[BIGLEN],
-			method[SMALLLEN];
+	char		method[SMALLLEN];
 	Inheader	*ih;
 
 	ih = inheadp;
@@ -427,45 +508,50 @@ char	*req;
 		while (  *cp && isspace( *cp))
 			cp++;
 	}
-	ih->protocol = ( *cp ? HTTP1_0 : HTTP0_9);
 
-	strcpy( ih->url_path, req); /* both arrays have size BIGLEN */
+	if ( !*cp )
+		ih->protocol =  HTTP0_9;
+	else if ( strncasecmp( cp, "HTTP/1.1", 8) == 0 )
+		ih->protocol =  HTTP1_1;
+	else
+		ih->protocol =  HTTP1_0;
+
+
+	mystrncpy( ih->url_path, req, BIGLEN);
 
 	if ( streq( method, "GET")) {
 		ih->method = GET;
-		if ( ih->protocol == HTTP1_0 ) 
-			get_header( ih);
+		if ( ih->protocol != HTTP0_9 ) 
+			get_headers( ih);
 		return	(ih->method);
 
 	}
 	if ( streq( method, "POST") ) {
 		if (serv_perm & WN_FORBID_EXEC) {
-			senderr( "501", errmsg, "");
+			senderr( "501", err_m[4], "");
 			return	(ih->method = UNKNOWN);
 		}
 		else {
-			get_header( ih);
+			get_headers( ih);
 			return	(ih->method = POST);
 		}
 	}
 	if ( streq( method, "PUT") ) {
 		if (serv_perm & WN_FORBID_EXEC) {
-			senderr( "501", errmsg, "");
+			senderr( "501", err_m[4], "");
 			return	(ih->method = UNKNOWN);
 		}
 		else {
-			get_header( ih);
+			get_headers( ih);
 			return	(ih->method = PUT);
 		}
 	}
 	if ( streq( method, "HEAD") ) {
-		get_header( ih);
+		get_headers( ih);
 		return	(ih->method = HEAD);
 	}
 	else {
-		strcpy( errmsg, method);
-		strcat( errmsg, ": Method not implemented");
-		senderr( "501", errmsg, "");
+		senderr( "501", err_m[116], "");
 		return	(ih->method = UNKNOWN);
 	}
 }
@@ -473,7 +559,7 @@ char	*req;
 
 
 static void
-get_header( ih)
+get_headers( ih)
 Inheader	*ih;
 {
 	register char	*cp, *cp2;
@@ -496,7 +582,7 @@ Inheader	*ih;
 				strcat( ih->accept, ", ");
 			acceptlen += strlen( cp);
 			if ( acceptlen > ACCEPTLEN)
-				logerr( ERRMSG19, "");
+				logerr( err_m[19], "");
 			else {
 				strcat( ih->accept, cp);
 			}
@@ -508,9 +594,9 @@ Inheader	*ih;
 			while ( isspace( *cp))
 				cp++;
 			if ( strlen( cp) > ACCEPTLEN/4 )
-				logerr( ERRMSG19, "");
+				logerr( err_m[19], "");
 			else {
-				strcpy( ih->lang, cp);
+				mystrncpy( ih->lang, cp, ACCEPTLEN/4);
 			}
 			continue;
 		}
@@ -520,9 +606,9 @@ Inheader	*ih;
 			while ( isspace( *cp))
 				cp++;
 			if ( strlen( cp) > ACCEPTLEN/4 )
-				logerr( ERRMSG19, "");
+				logerr( err_m[19], "");
 			else {
-				strcpy( ih->charset, cp);
+				mystrncpy( ih->charset, cp, ACCEPTLEN/4);
 			}
 			continue;
 		}
@@ -535,7 +621,7 @@ Inheader	*ih;
 				strcat( ih->cookie, "; ");
 			cookielen += strlen( cp);
 			if ( cookielen > ACCEPTLEN )
-				logerr( ERRMSG79, "");
+				logerr( err_m[79], "");
 			else {
 				strcat( ih->cookie, cp);
 			}
@@ -611,7 +697,7 @@ Inheader	*ih;
 			cp = headerline + 6;
 			while ( isspace( *cp))
 				cp++;
-			mystrncpy( ih->range,  cp, SMALLLEN);
+			mystrncpy( ih->range,  cp, RANGELEN);
 			continue;
 		}
 
@@ -623,6 +709,7 @@ Inheader	*ih;
 			continue;
 		}
 		if ( strncasecmp( headerline, "If-Modified-Since:", 18) == 0) {
+						/* it's a Fielding wart */
 			if ( ih->method == POST)
 				continue;
 			cp = headerline + 18;
@@ -630,20 +717,63 @@ Inheader	*ih;
 				cp++;
 			mystrncpy( ih->inmod_date, cp, SMALLLEN);
 			ih->method = CONDITIONAL_GET;
+			ih->conget |= IFMODSINCE;
 			continue;
 		}
-		if ( strncasecmp( headerline, "If-Unmodified-Since:", 22)
+
+		if ( strncasecmp( headerline, "If-Unmodified-Since:", 20)
 									== 0) {
-			if ( (ih->method == POST)
-					|| (ih->method == UNLESS_GET))
+						/* another Fielding wart */
+			if ( ih->method == POST)
 				continue;
-			cp = headerline + 22;
+			cp = headerline + 20;
 			while ( isspace( *cp))
 				cp++;
 			mystrncpy( ih->inmod_date, cp, SMALLLEN);
-			ih->method = UNLESS_GET;
+			ih->method = CONDITIONAL_GET;
+			ih->conget |= IFUNMODSINCE;
 			continue;
 		}
+
+		if ( strncasecmp( headerline, "If-None-Match:", 14) == 0) {
+			if ( ih->method == POST)
+				continue;
+			cp = headerline + 14;
+			while ( isspace( *cp))
+				cp++;
+			mystrncpy( ih->etag, cp, SMALLLEN);
+			ih->conget |= IFNMATCH;
+			ih->method = CONDITIONAL_GET;
+			continue;
+		}
+
+		if ( strncasecmp( headerline, "If-Match:", 9) == 0) {
+			if ( ih->method == POST)
+				continue;
+			cp = headerline + 9;
+			while ( isspace( *cp))
+				cp++;
+			mystrncpy( ih->etag, cp, SMALLLEN);
+			ih->conget |= IFMATCH;
+			ih->method = CONDITIONAL_GET;
+			continue;
+		}
+
+		if ( strncasecmp( headerline, "If-Range:", 9) == 0) {
+			if ( ih->method == POST)
+				continue;
+			cp = headerline + 9;
+			while ( isspace( *cp))
+				cp++;
+			if ( strchr( cp, ':') != NULL )   /* It's a date */
+				mystrncpy( ih->inmod_date, cp, SMALLLEN);
+			else	/* It's an ETag */
+				mystrncpy( ih->etag, cp, SMALLLEN);
+			ih->conget |= IFRANGE;
+			ih->method = CONDITIONAL_GET;
+			continue;
+		}
+
 	}
 }
 
@@ -652,28 +782,34 @@ static void
 do_post( ih)
 Inheader	*ih;
 {
+#ifndef FORBID_CGI
 	long	len;
 	int	c;
-	char	tfile[SMALLLEN];
 	FILE	*fp;
-
+	static int	num = 0;
 #ifndef STANDALONE
 	umask( 077);
 #endif
-	strcpy( tfile, TEMPDIR);
-	strcat( tfile, "/WNpostXXXXXX");
-	mktemp( tfile);
-	strcpy( ih->tmpfile, tfile);
+	len = (long) 0;
+	sprintf( ih->tmpfile, "%.200s/WNpost-%x-%x-%x",
+				wn_tmpdir, time( NULL), this_conp->pid, num++);
 
-	if ( (fp = fopen( tfile, "w")) == (FILE *) NULL ) {
-		senderr( SERV_ERR, ERRMSG53, tfile);
-		wn_exit( 2);
+	if ( (fp = fopen( ih->tmpfile, "w")) == (FILE *) NULL ) {
+		senderr( SERV_ERR, err_m[53], ih->tmpfile);
+		wn_exit( 2); /* senderr: SERV_ERR */
+	}
+
+
+	if ( *(ih->length))
+		len = atol( ih->length);
+	else {
+		senderr( "403", err_m[57], "");
+		wn_exit( 0);
 	}
-	len = atol( ih->length);
 
 	if ( (len > MAX_POST) || (len < 0)) {
-		senderr( "403", ERRMSG74, tfile);
-		wn_exit( 0);
+		senderr( "403", err_m[74], "");
+		wn_exit( 2);
 	}
 
 	/* Timeout after TRANSACTION_TIMEOUT seconds */
@@ -684,7 +820,7 @@ Inheader	*ih;
 		putc( c, fp);
 	}
 	fclose( fp);
-
+#endif
 }
 
 
@@ -703,10 +839,12 @@ wn_getc()
 		if ( n <= 0 )
 			return EOF;
 	}
-	c = (int) *(bp->bcp);
+	c = (unsigned char) *(bp->bcp);
 	(bp->bcp)++;
 	return c;
 }
+
+
 /*
  * static char *get_input()
  * Returns NULL if no input (client quit), otherwise places next line
@@ -716,6 +854,8 @@ wn_getc()
  * it.  It looks ahead to see if the line continues (i.e. next line starts
  * with white space.  The lookahead is only done for header lines
  * (rline = FALSE) not for the request line.
+ * The only possibilities for "line" are request and headerline.
+ * Both have size BIGLEN.
  */
 
 static char *
@@ -723,7 +863,7 @@ get_input( line, is_rline)
 char	*line;
 int	is_rline; 
 {
-	int	llen = 0,
+	int	llen,
 		n = 0;
 	Inbuffer	*bp;
 
@@ -734,15 +874,15 @@ int	is_rline;
 		bp->bcp = bp->buffer;
 
 	*line = '\0';
+	llen = 0;
 	while ( TRUE) {
-		if ( bp->cur_sz <= 0 ) {  /* fill the buffer */
+		if ( reset_buf( bp) == FALSE ) {  /* the buffer is empty */
 			n = load_inbuf( bp);
 			if ( n <= 0) {
 				if ( debug_log) {
 					char	tmpbuf[SMALLLEN];
-					sprintf( tmpbuf,
-					"PID %d got null or error read (%d) ",
-						this_conp->pid, n);
+					sprintf( tmpbuf, log_m[20],
+							this_conp->pid, n);
 					write_debug(1, tmpbuf, "");
 				}
 				return NULL;
@@ -752,6 +892,8 @@ int	is_rline;
 		if ( ( cp = strchr( bp->bcp, '\n')) == NULL) {
 			/* an incomplete header has been read */
 			/* put what there is into line */
+			/* if header contains nulls, stuff will be discarded */
+
 			bp->buffer[bp->cur_sz] = '\0';
 			mystrncat( line, bp->bcp, BIGLEN - llen);
 			llen = strlen( line);
@@ -762,24 +904,24 @@ int	is_rline;
 
 			n = load_inbuf( bp);
 			if ( n <= 0 ) {
-				writelog( this_rp,  ERRMSG66, line);
+				writelog( this_rp,  err_m[66], line);
 				wn_exit( 2);
 			}
 			continue;
 		}
 		
 		/* cp now points to next NL in the in_bufer */
-		cp--;
-		if ( ( cp >= bp->buffer) && *cp == '\r')
-			*cp = '\0';
-		cp++;
 		*cp++ = '\0';
 
 		mystrncpy( line + llen, bp->bcp, BIGLEN - llen);
 		llen = strlen( line);
+
+		if ( (llen > 0) && (line[llen-1] == '\r'))
+			line[--llen] = '\0';
+
 		bp->bcp = cp;
 
-		reset_buf (bp);
+		this_conp->more_in_buf = reset_buf(bp);
 
 		if ( debug_log)
 			write_debug(1, " -> ", line);
@@ -841,7 +983,7 @@ Inbuffer	*bp;
 	
 /*
  * static int chk_continue( bp)
- * Check if input line continues (next line starts with whitespace .
+ * Check if input line continues (next line starts with whitespace.
  * If buffer is empty, reinitialize it.
  */
 
@@ -869,20 +1011,24 @@ Inbuffer	*bp;
 /*
  * static void reset_buf( bp)
  * Check if buffer is empty and reset it.
+ * Return FALSE if empty and TRUE otherwise.
  */
 
-static void
+static int
 reset_buf( bp)
 Inbuffer	*bp;
 {
 
-	if ( bp->bcp - bp->buffer < bp->cur_sz)
-		return;  /* it's not empty yet */
-	else {
+	if ( bp->bcp - bp->buffer >= bp->cur_sz) {
+		/* it's empty */
 		bp->bcp = bp->buffer;
 		*(bp->bcp) = '\0';
 		bp->cur_sz = 0;
+		return FALSE;
 	}
+	else
+		return TRUE;  /* it's not empty yet */
+
 }
 
 
@@ -912,6 +1058,7 @@ http_prolog( )
 			datebuf[2*TINYLEN];
 	unsigned	unbuffered;
 
+
 	ip = this_rp;
 
 	if ( ip->status & WN_PROLOGSENT )
@@ -920,170 +1067,196 @@ http_prolog( )
 	unbuffered = ip->attributes & WN_UNBUFFERED;
 	ip->attributes &= ~(WN_UNBUFFERED);  /* always buffer HTTP headers */
 
-	if ( inheadp->protocol ==  HTTP1_0) {
-		if ( *outheadp->status)
-			sprintf( buf, "%s %s\r\n",
-					HTTPVERSION, outheadp->status);
-		else {
-			strcpy( buf, HTTPVERSION);
-			strcat( buf, " 200 OK\r\n");
-		}
-		send_text_line( buf);
-		sprintf( buf, "Server: %s\r\n", VERSION);
-		send_text_line( buf);
+	if ( inheadp->protocol ==  HTTP0_9) {
+		ip->attributes |= unbuffered;
+		return TRUE;
+	}
 
+	if ( *outheadp->status)
+		sprintf( buf, "%s %s\r\n",
+				HTTPVERSION, outheadp->status);
+	else {
+		strcpy( buf, HTTPVERSION);
+		strcat( buf, " 200 OK\r\n");
+	}
+	send_text_line( buf);
+	sprintf( buf, "Server: %s\r\n", VERSION);
+	send_text_line( buf);
 		/* Find date and serve the HTTP Date header */
-		time(&clock);
-		gmt = gmtime(&clock);
-		strftime( datebuf, SMALLLEN, 
-				"Date: %a, %d %h %Y %T GMT\r\n", gmt);
-		send_text_line( datebuf);
+	time(&clock);
+	gmt = gmtime(&clock);
+	strftime( datebuf, SMALLLEN, 
+			"Date: %a, %d %h %Y %T GMT\r\n", gmt);
+	send_text_line( datebuf);
+	if ( strncmp( outheadp->status, "204", 3) == 0 ) {
+		send_text_line( "Content-length: 0\r\n\r\n");
+		return TRUE;
+	}
 
-		if ( strncmp( outheadp->status, "204", 3) == 0 ) {
-			send_text_line( "Content-length: 0\r\n\r\n");
-			return TRUE;
-		}
+	if ( strncmp( outheadp->status, "401", 3) == 0 ) {
+		send_text_line( outheadp->list);
+		if ( this_conp->keepalive)
+			send_text_line( "Connection: Keep-Alive\r\n");
+		await_timeout = AUTHENT_TIMEOUT;
+		sprintf( buf, "Content-length: %s\r\n\r\n", ip->length);
+		send_text_line( buf);
+		return TRUE;
+	}
 
-		if ( strncmp( outheadp->status, "401", 3) == 0 ) {
-			send_text_line( outheadp->list);
-			if ( this_conp->keepalive)
-				send_text_line( "Connection: Keep-Alive\r\n");
-			await_timeout = AUTHENT_TIMEOUT;
-			sprintf( buf, "Content-length: %s\r\n\r\n",
-						ip->length);
-			send_text_line( buf);
-			return TRUE;
-		}
+	if ( *outheadp->redirect) {
+		send_text_line( "Content-type: text/html\r\n");
+		if ( this_conp->keepalive)
+			send_text_line( "Connection: Keep-Alive\r\n");
+		send_text_line( "Location: ");
+		send_text_line( outheadp->redirect);
+		send_text_line( "\r\nContent-length: 0\r\n\r\n");
+		return TRUE;
+	}
 
-		if ( *outheadp->redirect) {
-			send_text_line( "Content-type: text/html\r\n");
-			if ( this_conp->keepalive)
-				send_text_line( "Connection: Keep-Alive\r\n");
-			send_text_line( "Location: ");
-			send_text_line( outheadp->redirect);
-			send_text_line( "\r\nContent-length: 0\r\n\r\n");
-			return TRUE;
-		}
+	if ( ip->attributes & (WN_DYNAMIC)) {
+		ip->mod_time = (time_t) 0;
+		*ip->mod_date = '\0';
+		*ip->length = '\0';
+	}
 
-		if ( ip->attributes & (WN_DYNAMIC + WN_NOCACHE)) {
-			ip->mod_time = (time_t) 0;
-			*ip->mod_date = '\0';
-			*ip->length = '\0';
-		}
-
-			
-		if ( ip->mod_time) {
-			gmt = gmtime(&ip->mod_time);
-			strftime( ip->mod_date, SMALLLEN,
+	if ( ip->attributes & (WN_NOCACHE)) {
+		send_text_line( "Pragma: no-cache\r\n");
+		if ( inheadp->protocol ==  HTTP1_1)
+			send_text_line( "Cache-control: no-cache\r\n");
+	}
+		
+	if ( ip->mod_time) {
+		gmt = gmtime(&ip->mod_time);
+		strftime( ip->mod_date, SMALLLEN,
 				"Last-modified: %a, %d %h %Y %T GMT\r\n", gmt);
+
+#if 0
+		if ( (!(ip->attributes & (WN_FILTERED + WN_PARSE)))
+					((clock - ip->mod_time) > 2))  {
+			sprintf( buf, "ETag: \"%s\"\r\n", ip->etag);
+			send_text_line( buf);
 		}
+#endif
+	}
 
-		if ( *ip->mod_date) {
-			send_text_line( ip->mod_date);
-		}
+	if ( *ip->mod_date) {
+		send_text_line( ip->mod_date);
+	}
 
-		if ( ip->maxage && *ip->maxage) {
-			long	delta;
+	if ( ip->maxage && *ip->maxage) {
+		long	delta;
 
-			if ( *ip->maxage == 'L') {
-				ip->maxage++;
-				clock2 = ip->mod_time + atol( ip->maxage);
-				delta = clock2 - clock;
+		if ( *ip->maxage == 'L') {
+			ip->maxage++;
+			clock2 = ip->mod_time + atol( ip->maxage);
+			delta = clock2 - clock;
 
-				if ( delta > 0) {
-					sprintf( buf, 
-					"Cache-Control: max-age=%ld\r\n",
-								delta);
-					send_text_line( buf);
-				}
-			}
-			else {
-				clock2 = clock + atol( ip->maxage);
-				sprintf( buf, "Cache-Control: max-age=%s\r\n",
-						ip->maxage);
+			if ( delta > 0) {
+				sprintf( buf, "Cache-Control: max-age=%ld\r\n",
+							 delta);
 				send_text_line( buf);
 			}
 		}
-
-		if ( ip->expires && *ip->expires) {
-			sprintf( buf, "Expires: %s\r\n", ip->expires);
+		else {
+			clock2 = clock + atol( ip->maxage);
+			sprintf( buf, "Cache-Control: max-age=%s\r\n",
+					ip->maxage);
 			send_text_line( buf);
 		}
-		else if ( *outheadp->expires) {
-			send_text_line( outheadp->expires);
-		}
-		else if ( ip->maxage && *ip->maxage) {
-			gmt = gmtime(&clock2);
-			strftime( datebuf, SMALLLEN, 
-				"Expires: %a, %d %h %Y %T GMT\r\n", gmt);
-			send_text_line( datebuf);
-		}
-
-		if ( ip->maxage && *ip->maxage) {
-		}
-
-		if ( inheadp->method == CONDITIONAL_GET) {
-			/* It's not modified */
-			if ( this_conp->keepalive)
-				send_text_line( "Connection: Keep-Alive\r\n");
-			send_text_line( "\r\n");
-			return TRUE;
-		}
-
-		if ( ip->content_type) {
-			sprintf(buf, "Content-type: %s\r\n", ip->content_type);
-			send_text_line( buf);
-		}
-
-		if ( *ip->length && !(ip->attributes 
-				& (WN_PARSE + WN_FILTERED + WN_DYNAMIC))) {
-			sprintf( buf, "Content-length: %s\r\n", ip->length);
-			send_text_line( buf);
-
-			if ( *outheadp->range )
-				send_text_line( outheadp->range);
-
-			if ( !(ip->filetype & WN_BYTERANGE) &&
-					!*(outheadp->status)) {
-				/* !*(outheadp->status)) means status 200 */
-				send_text_line( "Accept-Ranges: bytes\r\n"); 
-			}
-		}
-		else if ( inheadp->method != HEAD)
-			this_conp->keepalive = FALSE;
-
-		if ( this_conp->keepalive) {
-			send_text_line( "Connection: Keep-Alive\r\n");
-		}
-				
-		if ( ip->encoding && *ip->encoding) {
-			sprintf( buf, "Content-encoding: %s\r\n",
-				ip->encoding);
-			send_text_line( buf);
-		}
-
-		if ( ip->title && *ip->title) {
-			sprintf( buf, "Title: %s\r\n", ip->title);
-			send_text_line( buf);
-		}
-
-		if ( ip->keywords && *ip->keywords) {
-			sprintf( buf, "Keywords: %s\r\n", ip->keywords);
-			send_text_line( buf);
-		}
-
-		if ( dir_p->dir_owner && *dir_p->dir_owner) {
-			sprintf( buf,"Link: <%s>; rev=\"Made\"\r\n",
-				dir_p->dir_owner);
-			send_text_line( buf);
-		}
-
-		if ( *outheadp->list ) {
-			send_text_line( outheadp->list);
-		}
-		send_text_line( "\r\n");
 	}
+
+	if ( ip->expires && *ip->expires) {
+		sprintf( buf, "Expires: %s\r\n", ip->expires);
+		send_text_line( buf);
+	}
+	else if ( *outheadp->expires) {
+		send_text_line( outheadp->expires);
+	}
+	else if ( ip->maxage && *ip->maxage) {
+		gmt = gmtime(&clock2);
+		strftime( datebuf, SMALLLEN, 
+			"Expires: %a, %d %h %Y %T GMT\r\n", gmt);
+		send_text_line( datebuf);
+	}
+
+	if ( inheadp->method == CONDITIONAL_GET) {
+		/* It's not modified */
+		if ( this_conp->keepalive)
+			send_text_line( "Connection: Keep-Alive\r\n");
+		send_text_line( "\r\n");
+		return TRUE;
+	}
+
+	if ( ip->content_type) {
+		sprintf(buf, "Content-type: %s\r\n", ip->content_type);
+		send_text_line( buf);
+	}
+
+	if ( *ip->length && !(ip->attributes 
+				& (WN_PARSE + WN_FILTERED + WN_DYNAMIC))) {
+		sprintf( buf, "Content-length: %s\r\n", ip->length);
+		send_text_line( buf);
+		if ( *outheadp->range )
+			send_text_line( outheadp->range);
+
+		if ( !(ip->filetype & WN_BYTERANGE) && !*(outheadp->status)) {
+			/* !*(outheadp->status)) means status 200 */
+			send_text_line( "Accept-Ranges: bytes\r\n"); 
+		}
+	}
+	else if ( inheadp->method != HEAD)
+		this_conp->keepalive = FALSE;
+
+	if ( this_conp->keepalive) {
+		send_text_line( "Connection: Keep-Alive\r\n");
+	}
+			
+	if ( ip->encoding && *ip->encoding) {
+		sprintf( buf, "Content-encoding: %s\r\n",
+			ip->encoding);
+		send_text_line( buf);
+	}
+
+#ifdef USE_TITLE_KEYWORD_HEADERS
+	if ( ip->title && *ip->title) {
+		sprintf( buf, "Content-description: %s\r\n", ip->title);
+		/* formally Title: */
+		send_text_line( buf);
+	}
+
+	if ( ip->keywords && *ip->keywords) {
+		sprintf( buf, "Keywords: %s\r\n", ip->keywords);
+		send_text_line( buf);
+	}
+#endif
+	if ( dir_p->dir_owner && *dir_p->dir_owner) {
+		sprintf( buf,"Link: <%s>; rev=\"Made\"\r\n",
+			dir_p->dir_owner);
+		send_text_line( buf);
+	}
+
+	if ( *outheadp->list ) {
+		send_text_line( outheadp->list);
+	}
+	send_text_line( "\r\n");
+
 	ip->attributes |= unbuffered;
 	return TRUE;
 }
 
+
+static void
+chk_method( meth)
+Methodtype meth;
+{
+
+	if ( inheadp->method != meth) {
+		if ( (meth == GET) && ( this_rp->attributes & WN_POST_OK) &&
+				((inheadp->method == POST) 
+				|| (inheadp->method == PUT)) )
+			return;		/* PUT or POST ok */
+
+		senderr( "405", err_m[117], "");
+		wn_exit( 2);  /* senderr: 405 method not allowed */
+	}
+}
