@@ -1,9 +1,9 @@
 /*
     Wn: A Server for the HTTP
     File: wndex/wndex.c
-    Version 1.17.0
+    Version 1.18.5
     
-    Copyright (C) 1995, 1996  <by John Franks>
+    Copyright (C) 1995-7  <by John Franks>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -50,6 +50,8 @@ static void	mkcache(),
 		despace(),
 		clear();
 
+static int	is_filedir();
+
 Entry		top;
 
 
@@ -72,7 +74,7 @@ static void
 clear( ep)
 Entry	*ep;
 {
-	ep->content[0] = ep->cacheline[0] = ep->url[0]
+	ep->content[0] = ep->cacheline[0] = ep->url[0] = ep->subdirs[0]
 			= ep->file[0] = ep->title[0] = '\0';
 	ep->flag = ep->attributes = 0;
 	ep->isindexfile = ep->foundtitle = ep->foundkey 
@@ -96,6 +98,8 @@ Entry	*ep;
 		tmpfpath[MIDLEN],
 		htmlpath[MIDLEN];
 
+	int	use_cntrlf = TRUE;
+
 
 	ep->owner[0] = '\0';
 	ep->firsttime = TRUE;
@@ -114,8 +118,18 @@ Entry	*ep;
 	if ( stdioflg)
 		cntlfp = stdin;
 	else if ( (cntlfp = fopen( ep->cntlfpath, "r")) == (FILE *) NULL ) {
-		fprintf( stderr, ERRMSG2, ep->cntlfpath); /* can't open */
-		return;
+		/* can't open */
+		if ( strong_serveall) {
+			use_cntrlf = FALSE;
+			if ( !ep->serveall) {
+				addpair( "serveall", "true", ep);
+				ep->serveall = TRUE;
+			}
+		}
+		else {
+			fprintf( stderr, ERRMSG2, ep->cntlfpath); 
+			return;
+		}
 	}
 
 	if ( stdioflg)
@@ -125,8 +139,8 @@ Entry	*ep;
 		return;
 	}
 
-	while ( bufp = get_next_line(mbuf, cntlfp)) {
-		cp = strchr( bufp, '=');
+	while ( use_cntrlf && (bufp = get_next_line(mbuf, cntlfp))) {
+		cp = strchr( bufp, '=');  /* guaranteed non-NULL */
 		*cp++ = '\0';
 		text = cp;
 		strlower( bufp);
@@ -268,8 +282,10 @@ Entry	*ep;
 			continue;
 		}
 		if ( streq( bufp, "accessfile")) {
-			if ( ep->firsttime)
+			if ( ep->firsttime) {
+				mystrncpy( ep->accessfile, text, SMALLLEN);
 				addpair( "accessfile", text, ep);
+			}
 			else
 				onlyfirst( bufp);
 			continue;
@@ -286,21 +302,30 @@ Entry	*ep;
 		}
 
 		if ( streq( bufp, "subdirs")) {
-			if ( ep->firsttime)
-				addpair( "subdirs", text, ep);
-			else
+			if ( !ep->firsttime) {
 				onlyfirst( bufp);
+				continue;
+			}
+
+			if ( streq( text, "<all>")) {
+				which_subdirs = WNDEX_ALL;
+				continue;
+			}
+
+			if ( streq( text, "<index>")) {
+				which_subdirs = WNDEX_INDEX;
+				continue;
+			}
+
+			addpair( "subdirs", text, ep);
+			mystrncpy( ep->subdirs, text, CACHELINE_LEN);
 
-			if ( recurse)
-				do_recurse( text, ep);
 			continue;
 		}
 
 		if ( streq( bufp, "title")) {
-			if ( ep->firsttime) {
-				fprintf( stderr, ERRMSG23);
-				exit( 2);
-			}
+			if ( !is_filedir( bufp, ep))
+				continue;
 			addpair( "title", text, ep);
 			strcpy( ep->title, text);
 			ep->foundtitle = TRUE;
@@ -308,11 +333,15 @@ Entry	*ep;
 		}
 
 		if ( streq( bufp, "header")) {
+			if ( !is_filedir( bufp, ep))
+				continue;
 			addpair( "header", text, ep);
 			continue;
 		}
 
 		if ( streq( bufp, "refresh")) {
+			if ( !is_filedir( bufp, ep))
+				continue;
 			strcpy( buf, "Refresh: ");
 			strcat( buf, text);
 			addpair( "header", buf, ep);
@@ -320,6 +349,8 @@ Entry	*ep;
 		}
 
 		if ( streq( bufp, "setcookie")) {
+			if ( !is_filedir( bufp, ep))
+				continue;
 			strcpy( buf, "Set-Cookie: ");
 			strcat( buf, text);
 			addpair( "header", buf, ep);
@@ -327,27 +358,37 @@ Entry	*ep;
 		}
 
 		if ( streq( bufp, "parse")) {
+			if ( !is_filedir( bufp, ep))
+				continue;
 			ep->attributes |= WN_PARSE;
 			continue;
 		}
 
 		if ( streq( bufp, "redirect")) {
+			if ( !is_filedir( bufp, ep))
+				continue;
 			addpair( "redirect", text, ep);
 			continue;
 		}
 
 		if ( streq( bufp, "keywords")) {
+			if ( !is_filedir( bufp, ep))
+				continue;
 			ep->foundkey = TRUE;
 			addpair( "keywords", text, ep);
 			continue;
 		}
 
 		if ( strncmp( bufp, "field", 5) == 0) {
+			if ( !is_filedir( bufp, ep))
+				continue;
 			addpair( bufp, text, ep);
 			continue;
 		}
 
 		if ( streq( bufp, "contenttype")) {
+			if ( !is_filedir( bufp, ep))
+				continue;
 			addpair( "content", text, ep);
 			strcpy( ep->content, text);
 			strlower( ep->content);
@@ -357,6 +398,10 @@ Entry	*ep;
 
 		if ( streq( bufp, "contentencoding")
 			|| streq( bufp, "encoding")) {
+
+			if ( !is_filedir( bufp, ep))
+				continue;
+
 			if ( !streq( text, "none")) {
 				addpair( "encoding", text, ep);
 			}
@@ -365,20 +410,26 @@ Entry	*ep;
 		}
 
 		if ( streq( bufp, "includes")) {
+			if ( !is_filedir( bufp, ep))
+				continue;
 			despace( text);
 			addpair( "includes", text, ep);
-			ep->attributes |= WN_INCLUDE;
+			continue;
+		}
+
+		if ( streq( bufp, "listincludes")) {
+			if ( !is_filedir( bufp, ep))
+				continue;
+			despace( text);
+			addpair( "list", text, ep);
 			continue;
 		}
 
 		if ( strncmp( bufp, "wrapper", 7) == 0 ) {
+			if ( !is_filedir( bufp, ep))
+				continue;
 			despace( text);
-			if ( ep->firsttime)
-				addpair( "dwrapper", text, ep);
-			else {
-				ep->attributes |= WN_WRAPPED;
-				addpair( "wrappers", text, ep);
-			}
+			addpair( "wrappers", text, ep);
 			continue;
 		}
 
@@ -386,13 +437,14 @@ Entry	*ep;
 			if ( ep->firsttime)
 				addpair( "dwrapper", text, ep);
 			else {
-				ep->attributes |= WN_SWRAPPED;
 				addpair( "swrapper", text, ep);
 			}
 			continue;
 		}
 
 		if ( streq( bufp, "emptysub")) {
+			if ( !is_filedir( bufp, ep))
+				continue;
 			addpair( "nomatchsub", text, ep);
 			continue;
 		}
@@ -403,12 +455,24 @@ Entry	*ep;
 		}
 
 		if ( streq( bufp, "filter")) {
+			if ( !is_filedir( bufp, ep))
+				continue;
 			ep->attributes |= WN_FILTERED;
 			addpair( "filter", text, ep);
 			continue;
 		}
 
+		if ( streq( bufp, "cgihandler")) {
+			if ( !is_filedir( bufp, ep))
+				continue;
+			ep->attributes |= WN_CGI;
+			addpair( "handler", text, ep);
+			continue;
+		}
+
 		if ( streq( bufp, "expires")) {
+			if ( !is_filedir( bufp, ep))
+				continue;
 			addpair( "expires", text, ep);
 			ep->foundexp = TRUE;
 			continue;
@@ -416,12 +480,16 @@ Entry	*ep;
 		if ( streq( bufp, "maxage")) {
 			char	minbuf[TINYLEN];
 
+			if ( !is_filedir( bufp, ep))
+				continue;
 			do_maxage( text, minbuf);
 			addpair( "maxage", minbuf, ep);
 			ep->foundexp = TRUE;
 			continue;
 		}
 		if ( streq( bufp, "nosearch")) {
+			if ( !is_filedir( bufp, ep))
+				continue;
 			ep->attributes |= WN_NOSEARCH;
 			continue;
 		}
@@ -438,11 +506,9 @@ Entry	*ep;
 			continue;
 		}
 
-		if ( streq( bufp, "defaultwrapper")) {
+		if ( strncmp( bufp, "defaultwrapper", 14) == 0 ) {
 			/* must go to server because of serveall */
-			strlower( text);
 			if ( ep->firsttime) {
-				strcpy( ep->defwrapper, text);
 				addpair( "defwrapper", text, ep);
 			}
 			else
@@ -452,9 +518,7 @@ Entry	*ep;
 
 		if ( streq( bufp, "defaultincludes")) {
 			/* must go to server because of serveall */
-			strlower( text);
 			if ( ep->firsttime) {
-				strcpy( ep->defincludes, text);
 				addpair( "defincludes", text, ep);
 			}
 			else
@@ -462,6 +526,16 @@ Entry	*ep;
 			continue;
 		}
 
+		if ( streq( bufp, "defaultlistincludes")) {
+			/* must go to server because of serveall */
+			if ( ep->firsttime) {
+				addpair( "deflist", text, ep);
+			}
+			else
+				onlyfirst( bufp);
+			continue;
+		}
+
 		if ( streq( bufp, "defaultdocument")) {
 			if ( ep->firsttime)
 				addpair( "default_document", text, ep);
@@ -470,6 +544,22 @@ Entry	*ep;
 			continue;
 		}
 
+		if ( streq( bufp, "defaultcgihandler")) {
+			if ( ep->firsttime)
+				addpair( "default_handler", text, ep);
+			else
+				onlyfirst( bufp);
+			continue;
+		}
+
+		if ( streq( bufp, "defaultfilter")) {
+			if ( ep->firsttime)
+				addpair( "default_filter", text, ep);
+			else
+				onlyfirst( bufp);
+			continue;
+		}
+
 		if ( streq( bufp, "defaultmaxage")) {
 			if ( ep->firsttime) {
 				char	minbuf[TINYLEN];
@@ -527,10 +617,11 @@ Entry	*ep;
 		do_serveall( buf, cachefp, htmlfp, ep);
 	}
 
-	fclose( cntlfp);
+	if ( use_cntrlf)
+		fclose( cntlfp);
 	fclose( cachefp);
 
-	if ( rename( tmpfpath, ep->cachefpath) != 0 ) {
+	if ( !stdioflg && (rename( tmpfpath, ep->cachefpath) != 0) ) {
 		fprintf( stderr, ERRMSG25, tmpfpath, ep->cachefpath);
 		return;
 	}
@@ -571,6 +662,10 @@ Entry	*ep;
 			cp2++;
 		}
 		currsub = cp;
+
+		if ( !*cp || streq( cp, ".") || streq(cp, ".."))
+			continue;
+
 		strcpy( next.cntlfpath, ep->cntlfpath);
 		if ( (cp = strrchr( next.cntlfpath, '/')) == NULL ) {
 			fprintf( stderr, ERRMSG6, ep->cntlfpath);
@@ -601,6 +696,19 @@ char	*s;
 	fprintf( stderr, ERRMSG8, s);
 }
 
+static int
+is_filedir( s, ep)
+char	*s;
+Entry	*ep;
+{
+	if ( ep->firsttime) {
+		if ( !quiet)
+			fprintf( stderr, ERRMSG23, s);
+		return FALSE;
+	}
+	return TRUE;
+}
+
 
 
 
@@ -643,9 +751,22 @@ Entry	*ep;
 		if ( *ep->file)
 			fprintf( hfp, "<li> <a href=\"%s\">%s</a>\n",
 				ep->file, buf);
-		if ( *ep->url)
+		if ( *ep->url) {
+			char	*cp,
+				buf2[MIDLEN];
+
+			mystrncpy( buf2, ep->url, MIDLEN);
+			cp = buf2;
+			cp++;
+			while ( *cp ) {
+				if ( (*cp == '&') && ( *(cp -1) == '\\'))
+					mystrncpy( cp - 1, cp, MIDLEN);
+				else
+					cp++;
+			}
 			fprintf( hfp, "<li> <a href=\"%s\">%s</a>\n",
-				ep->url, buf);
+						buf2, buf);
+		}
 	}
 	clear( ep);
 }
@@ -656,7 +777,22 @@ writedir( fp, ep)
 FILE	*fp;
 Entry	*ep;
 {
-	char	buf[MIDLEN];
+	char	buf[MIDLEN],
+		*cp;
+
+	if ( (which_subdirs != WNDEX_NONE) && !*(ep->subdirs)) {
+		strcpy( buf, ep->cachefpath);
+		if ( cp = strrchr( buf, '/'))
+			*cp = '\0';
+		mksubd_list( buf, ep);
+		if ( *(ep->subdirs))
+			addpair( "subdirs", ep->subdirs, ep);
+	}
+
+	if ( recurse) {
+		do_recurse( ep->subdirs, ep);
+	}
+
 
 	if ( ep->defattributes) {
 		sprintf( buf, "%ld", ep->defattributes);
@@ -772,7 +908,7 @@ Entry	*ep;
 /* 
  * Read in line, skip lines with no "=" in them, deal with comments (#),
  * get rid of leading and trailing whitespace.  If line ends with '\' 
- * it continues on next line.  Maximum allowed size of a line is BIGLEN.
+ * it continues on next line.  Maximum allowed size of a line is MIDLEN/2.
  */
 			
 char
@@ -780,23 +916,40 @@ char
 char	*buf;
 FILE	*fp;
 {
-	register char	*cp;
+	register char	*cp,
+			*cp2;
 
 	char	*bufp,
-		extrabuf[BIGLEN];
+		extrabuf[MIDLEN];
 
-	while ( bufp = fgets( buf, BIGLEN, fp)) {
+	while ( bufp = fgets( buf, MIDLEN/2, fp)) {
 		chop( bufp);
+
+		if ( (cp = strchr( bufp, '#')) != NULL) {
+			if ( ( cp != bufp) && (*(cp-1) == '\\'))
+				strcpy( cp-1, cp);
+			else
+				*cp = '\0';
+		}
+
 		cp = buf + strlen(buf) - 1;
 		while ( isspace( *cp) && ( cp >= buf))
 			*cp-- = '\0';	/* remove trailing whitespace */
 
 		while ( *cp == '\\') {
 			*cp = '\0';
-			fgets( extrabuf, BIGLEN, fp);
+			fgets( extrabuf, MIDLEN/2, fp);
 			chop( extrabuf);
+
+			if ( (cp = strchr( extrabuf, '#')) != NULL) {
+				if ( ( cp != extrabuf) && (*(cp-1) == '\\'))
+					strcpy( cp-1, cp);
+				else
+					*cp = '\0';
+			}
+
 			if ( strlen( buf) + strlen( extrabuf) 
-						> CACHELINE_LEN - 20 ) {
+						> MIDLEN/2 ) {
 				fprintf( stderr, ERRMSG1, buf);
 				exit( 2);
 			}
@@ -806,12 +959,7 @@ FILE	*fp;
 			strcat( buf, cp);
 			cp = buf + strlen(buf) - 1;
 		}
-		if ( (cp = strchr( bufp, '#')) != NULL) {
-			if ( ( cp != bufp) && (*(cp-1) == '\\'))
-				strcpy( cp-1, cp);
-			else
-				*cp = '\0';
-		}
+
 		while ( (*bufp == '\t') || (*bufp == ' '))
 			bufp++;
 		if ( (cp = strchr( bufp, '=')) == NULL) {
@@ -821,6 +969,26 @@ FILE	*fp;
 		}
 		break;
 	}
+
+	if ( bufp == NULL )
+		return NULL;
+	if ( (cp2 = cp = strchr( bufp, '=')) == NULL) {
+		fprintf( stderr, "Unknown internal error\n");
+		exit( 1);
+	}
+
+	/* Delete white space around first '=' sign */
+	cp--;
+	while ( isspace( *cp) && (cp > bufp))
+		cp--;
+	mystrncpy( ++cp, cp2, MIDLEN);
+	cp2 = cp;
+	cp++;
+	while ( *cp && isspace( *cp))
+		cp++;
+	mystrncpy( ++cp2, cp, MIDLEN);
+
+
 	return bufp;
 }
 	
@@ -1031,10 +1199,22 @@ unsigned	*attrib;
 			*attrib |= WN_NONDYNAMIC;
 			continue;
 		}
+		if ( strncmp( word, "cach", 4) == 0 ) {
+			*attrib |= WN_CACHEABLE;
+			continue;
+		}
+		if ( strncmp( word, "noncach", 7) == 0 ) {
+			*attrib |= WN_NOCACHE;
+			continue;
+		}
 		if ( streq( word, "parse") ) {
 			*attrib |= WN_PARSE;
 			continue;
 		}
+		if ( streq( word, "post") ) {
+			*attrib |= WN_POST_OK;
+			continue;
+		}
 		if ( streq( word, "noparse") ) {
 			*attrib |= WN_NOPARSE;
 			continue;
@@ -1069,9 +1249,14 @@ Entry		*ep;
 
 		if ( streq( word, "serveall") ) {
 #ifndef NO_SERVEALL
-			addpair( "serveall", "true", ep);
+			if ( !ep->serveall) {
+				addpair( "serveall", "true", ep);
+				ep->serveall = TRUE;
+			}
+#else
+			if ( !quiet)
+				fprintf( stderr, ERRMSG31);
 #endif
-			ep->serveall = TRUE;
 			continue;
 		}
 		if ( streq( word, "nosearch") ) {
